// pub mod group;
use crate::bits::*;
use crate::group::*;

use std::string::String;
use strum::IntoEnumIterator;
extern crate rand_chacha;
use rand::distributions::{Distribution, Uniform};
use rand::seq::SliceRandom;
use rand_chacha::rand_core::SeedableRng;
use rand_chacha::ChaCha20Rng;

// For custom error
use std::error;
use std::fmt;

#[derive(Debug, Clone)]
struct GameLoadError {
    message: String,
}

impl fmt::Display for GameLoadError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Game load error: {}", self.message)
    }
}

impl error::Error for GameLoadError {}

// const DEBUG_OUTPUT: bool = false;

// Vec doesn't implement Copy ...
#[derive(Debug, Clone, Eq, PartialEq)]
pub struct AnyBoard {
    pub size: u8,
    pub width: u8,
    pub max_index: usize,
    pub board: Vec<u8>,
}

impl AnyBoard {
    pub fn new(board_size: u8) -> Self {
        if (board_size < 3) || (board_size > 5) {
            panic!("Board size must be 3-5.");
        }

        let n = board_size as usize;
        let s = AnyBoard {
            size: board_size,
            width: board_size * board_size,
            max_index: n * n * n * n,
            board: vec![0; n * n * n * n],
        };
        s
    }

    /// Clear out the board
    pub fn clear(&mut self) {
        self.board.fill(0);
    }

    pub fn copy(&mut self, copy_from: &Self) {
        debug_assert!(
            self.size == copy_from.size,
            "Can't copy size {} into size {}",
            copy_from.size,
            self.size
        );
        for i in 0..self.max_index {
            self.board[i] = copy_from.board[i];
        }
    }

    #[inline]
    #[must_use]
    /// Calculate index position of (x,y)
    pub fn pos(&self, x: u8, y: u8) -> usize {
        debug_assert!(
            x < self.width && y < self.width,
            "Expected ({}, {}) < {}",
            x,
            y,
            self.width
        );
        x as usize + y as usize * self.width as usize
    }

    #[must_use]
    #[inline]
    /// Return (x,y) position for given index.
    pub fn xy(&self, idx: usize) -> (u8, u8) {
        (
            (idx % self.width as usize) as u8,
            (idx / self.width as usize) as u8,
        )
    }

    /// Set a position in the board with a value
    pub fn set(&mut self, x: u8, y: u8, value: u8) {
        debug_assert!(
            x < self.width && y < self.width,
            "Expected ({}, {}) < {}",
            x,
            y,
            self.width
        );
        debug_assert!(
            value <= self.width,
            "Expected value for ({},{}) = {} < {}",
            x,
            y,
            value,
            self.width
        );
        let index = self.pos(x, y);
        assert!(index <= self.board.capacity());
        self.board[index] = value;
    }

    #[must_use]
    /// Get value at position (x,y)
    pub fn get(&self, x: u8, y: u8) -> u8 {
        debug_assert!(
            x < self.width && y < self.width,
            "Expected ({}, {}) < {}",
            x,
            y,
            self.width
        );
        let index = self.pos(x, y);
        assert!(index <= self.board.capacity());
        self.board[index]
    }

    #[must_use]
    /// Load from ksudoku format
    /// - uses load_from_tld with specifics
    pub fn load_ksudoku(&mut self, s: &str) -> Result<(), Box<dyn error::Error>> {
        self.load_from_tld('b', '_', s)
    }

    #[must_use]
    /// Save to ksudoku format
    pub fn save_ksudoku(&self) -> String {
        self.save_to_tld('b', '_')
    }

    #[must_use]
    /// Load puzzle from string (top,left) going down.
    pub fn load_from_tld(
        &mut self,
        start_ch: char,
        blank: char,
        s: &str,
    ) -> Result<(), Box<dyn error::Error>> {
        self.clear();
        let mut x: u8 = 0;
        let mut y: u8 = 0;

        if s.len() != self.max_index {
            // self.size * self.size*self.size*self.size {
            return Err(Box::new(GameLoadError {
                message: format!(
                    "String ({}) exceeds expected length {}.",
                    s.len(),
                    self.width
                ),
            }));
        }

        for ch in s.chars() {
            if ch != blank {
                let value = (ch as u8 - start_ch as u8) + 1;
                if value == 0 || value > self.width {
                    return Err(Box::new(GameLoadError {
                        message: format!(
                            "String symbol ({}) represents value {}, expecting 1 to {}.",
                            ch, value, self.width
                        ),
                    }));
                }
                self.set(x, y, value);
            }
            y += 1;
            if y >= self.width {
                y = 0;
                x += 1;
            }
        }
        Ok(())
    }

    #[must_use]
    /// Save puzzle to a string (top,left) going down.
    pub fn save_to_tld(&self, start_ch: char, blank: char) -> String {
        let mut result = String::new();
        result.reserve(self.max_index);
        let start_ch = (start_ch as u8 - 1) as char;
        let mut x: u8 = 0;
        let mut y: u8 = 0;

        for _i in 0..self.max_index {
            let value = self.get(x, y);
            if value == 0 {
                result.push(blank);
            } else {
                result.push((start_ch as u8 + value) as char);
            }
            y += 1;
            if y >= self.width {
                y = 0;
                x += 1;
            }
        }
        result
    }

    #[must_use]
    /// Load puzzle from string (top,left) going right.
    pub fn load_from_tlr(
        &mut self,
        start_ch: char,
        blank: char,
        s: &str,
    ) -> Result<(), Box<dyn error::Error>> {
        self.clear();

        if s.len() != self.max_index {
            // self.size * self.size*self.size*self.size {
            return Err(Box::new(GameLoadError {
                message: format!(
                    "String exceeds ({}) expected length {}.",
                    s.len(),
                    self.width
                ),
            }));
        }

        let mut i: usize = 0;

        for ch in s.chars() {
            if ch != blank {
                let value = (ch as u8 - start_ch as u8) + 1;
                if value == 0 || value > self.width {
                    return Err(Box::new(GameLoadError {
                        message: format!(
                            "String symbol ({}) represents value {}, expecting 1 to {}.",
                            ch, value, self.width
                        ),
                    }));
                }
                self.board[i] = value;
                i += 1;
            }
        }
        Ok(())
    }

    #[must_use]
    /// Save puzzle to a string (top,left) going right.
    pub fn save_to_tlr(&self, start_ch: char, blank: char) -> String {
        let mut result = String::new();
        result.reserve(self.max_index);
        let start_ch = (start_ch as u8 - 1) as char;

        for i in 0..self.max_index {
            let value = self.board[i];
            if value == 0 {
                result.push(blank);
            } else {
                result.push((start_ch as u8 + value) as char);
            }
        }
        result
    }

    /// Display board using unicode box characters.
    pub fn display(&self) {
        let line = "═".repeat(self.size as usize);
        let alpha_display = self.width >= 10;
        // println!("╔{}╦{}╦{}╗", line, line, line);

        // Top
        for i in 0..self.size {
            if i == 0 {
                print!("╔{}", line);
            } else {
                print!("╦{}", line);
            }
        }
        println!("╗");

        for y in 0..self.width {
            print!("║");
            for x in 0..self.width {
                let value = self.get(x, y);
                if value == 0 {
                    print!(" ");
                } else {
                    if alpha_display {
                        print!("{}", (value - 1 + 'A' as u8) as char);
                    } else {
                        print!("{}", value);
                    }
                }
                if x % self.size == self.size - 1 {
                    print!("║");
                }
            }
            println!("");
            if y % self.size == self.size - 1 {
                if y + 1 == self.width {
                    // Bottom
                    for i in 0..self.size {
                        if i == 0 {
                            print!("╚{}", line);
                        } else {
                            print!("╩{}", line);
                        }
                    }
                    println!("╝");

                    // println("╚═══╩═══╩═══╝");
                } else {
                    // Middle
                    for i in 0..self.size {
                        if i == 0 {
                            print!("╠{}", line);
                        } else {
                            print!("╬{}", line);
                        }
                    }
                    println!("╣");
                    // println("╠═══╬═══╬═══╣");
                }
            }
        }
    }

    #[must_use]
    /// Output board as strings.
    /// - Uses 1-9 (for 9x9), and A-? for others.
    /// - Used by tests to confirm the board is what we think it should be.
    pub fn to_strings(&self) -> Vec<String> {
        let mut result = Vec::<String>::new();
        let alpha_display = self.width >= 10;

        for y in 0..self.width {
            let mut line = String::new();
            line.reserve(self.width as usize);
            for x in 0..self.width {
                let value = self.get(x, y);
                if value == 0 {
                    line.push(' ');
                } else {
                    if alpha_display {
                        line.push((value - 1 + 'A' as u8) as char);
                    } else {
                        line.push((value + '0' as u8) as char);
                    }
                }
            }
            result.push(line);
        }
        result
    }

    /// Is the puzzle completed?
    /// Have all of the locations been filled with a value?
    /// - This does not validate that it is a correct puzzle.
    ///   It doesn't check for duplicate digits (for example).
    pub fn complete(&self) -> bool {
        for i in 0..self.max_index {
            if self.board[i] == 0 {
                return false;
            }
        }
        true
    }

    #[must_use]
    /// Solve by brute force
    /// - Returns up to max # of solutions.
    /// - Returns total solutions found, and vector of boards (up to max).
    /// - Uses brute_force (recursive function) to solve.
    pub fn brute_force_solver(&self, max: u16) -> (u16, Vec<AnyBoard>) {
        let mut workset = self.clone();
        let mut total_solutions: u16 = 0;
        let mut solutions: Vec<AnyBoard> = Vec::new();
        let groups = AnyGroup::new(workset.size);
        solutions.reserve(max as usize);

        workset.brute_force(&mut total_solutions, &mut solutions, &groups);
        (total_solutions, solutions)
    }

    /// Recursive brute force solver.
    fn brute_force(
        &mut self,
        total_solutions: &mut u16,
        solutions: &mut Vec<AnyBoard>,
        groups: &AnyGroup,
    ) -> bool {
        for idx in 0..self.max_index {
            if self.board[idx] == 0 {
                // Blank found
                let (x, y) = self.xy(idx);

                // println!("Blank ({},{})", x, y);

                'outer: for value in 1..=self.width {
                    // Check if it fits in the puzzle via group.
                    for clear in groups.row(y) {
                        if self.board[*clear] == value {
                            continue 'outer;
                        }
                    }
                    for clear in groups.column(x) {
                        if self.board[*clear] == value {
                            continue 'outer;
                        }
                    }
                    for clear in groups.cell(groups.which_cell(x, y)) {
                        if self.board[*clear] == value {
                            continue 'outer;
                        }
                    }

                    // Ok, this is possible move.
                    self.board[idx] = value;
                    // println!("Try ({},{}) = {}", x, y, value);
                    // self.display();

                    if self.complete() {
                        if *total_solutions < solutions.capacity() as u16 {
                            solutions.push(self.clone());
                        }
                        *total_solutions += 1;
                        break;
                    } else {
                        if self.brute_force(total_solutions, solutions, groups) {
                            return true;
                        }
                    }
                }
                // We failed to place a value -- return failure.
                // println!("rewind");
                self.board[idx] = 0;
                return false;
            }
        }
        false
    }
}

// Ok! This is interesting:

use std::ops::{Index, IndexMut};

impl Index<usize> for AnyBoard {
    type Output = u8;

    fn index<'a>(&'a self, i: usize) -> &'a u8 {
        &self.board[i]
    }
}

impl Index<(u8, u8)> for AnyBoard {
    type Output = u8;

    fn index<'a>(&'a self, i: (u8, u8)) -> &'a u8 {
        &self.board[self.pos(i.0, i.1)]
    }
}

impl IndexMut<usize> for AnyBoard {
    fn index_mut<'a>(&'a mut self, i: usize) -> &'a mut u8 {
        &mut self.board[i]
    }
}

impl IndexMut<(u8, u8)> for AnyBoard {
    fn index_mut<'a>(&'a mut self, i: (u8, u8)) -> &'a mut u8 {
        let idx = self.pos(i.0, i.1);
        &mut self.board[idx]
    }
}

// Need to use u32, so 5*5=25, 25 bits can be accessed.
// u16 is fine for 3*3=9.

#[derive(Debug, Clone)]
pub struct AnyPossible {
    pub size: u8,
    pub width: u8,
    pub max_index: usize,
    pub possible: Vec<GenBits>,
}

impl AnyPossible {
    pub fn new(board_size: u8) -> Self {
        let width = board_size as usize * board_size as usize;
        let mut initial = GenBits::new(width);
        // let width = board_size * board_size;
        initial.set_bits(0..width as usize);

        Self {
            size: board_size,
            width: width as u8,
            max_index: width * width,
            possible: vec![initial; width * width],
        }
    }

    pub fn clear(&mut self) {
        let mut initial = GenBits::new(self.width as usize);
        // let width = self.size * self.size;
        initial.set_bits(0..self.width as usize);
        self.possible.fill(initial);
    }

    pub fn copy(&mut self, copy_from: &Self) {
        debug_assert!(
            self.size == copy_from.size,
            "Can't copy size {} into size {}",
            copy_from.size,
            self.size
        );
        for i in 0..self.max_index {
            self.possible[i] = copy_from.possible[i].clone();
        }
    }

    // NOTE: values range from 1..width. Bits range from 0..width-1

    #[inline]
    /// Set value's state for given index.
    ///   - Use value+1 ?
    ///   - value range is from 1..=width.
    ///   - return true if state changed.
    pub fn set(&mut self, index: usize, value: usize, state: bool) -> bool {
        debug_assert!(
            index < self.max_index,
            "Index {} >= {}",
            index,
            self.max_index
        );
        let changed = self.possible[index].get(value - 1) != state;
        self.possible[index].set(value - 1, state);
        changed
    }

    #[must_use]
    #[inline]
    /// Get state for given index and value.
    ///   - value range is from 1..=width.
    pub fn get(&self, index: usize, value: usize) -> bool {
        debug_assert!(
            index < self.max_index,
            "Index {} >= {}",
            index,
            self.max_index
        );
        self.possible[index].get(value - 1)
    }

    #[must_use]
    #[inline]
    /// Return index for given (x,y).
    pub fn pos(&self, x: u8, y: u8) -> usize {
        debug_assert!(
            x < self.width && y < self.width,
            "Expected ({},{}) < {}",
            x,
            y,
            self.width
        );
        x as usize + y as usize * self.width as usize
    }

    #[must_use]
    #[inline]
    /// Return (x,y) position for given index.
    pub fn xy(&self, idx: usize) -> (u8, u8) {
        debug_assert!(idx < self.max_index, "Index {} >= {}", idx, self.max_index);
        (
            (idx % self.width as usize) as u8,
            (idx / self.width as usize) as u8,
        )
    }

    #[must_use]
    /// Format all possible, finding max length.
    /// - Return vec<string> of each item formatted.
    /// - Return max length.
    pub fn pos_max(&self) -> (Vec<String>, usize) {
        let mut pos = Vec::<String>::new();
        pos.reserve(self.max_index);
        let mut max: usize = 0;

        for idx in 0..self.max_index {
            let s = self.possible[idx].display();
            if max < s.len() {
                max = s.len();
            }
            pos.push(s);
        }
        (pos, max)
    }

    #[must_use]
    /// find pairs (same possible number in different cell locations)
    /// This finds them all, and is not limited to finding 2.
    /// Returns
    ///  - Vector of (Vec(positions), Vec(values))
    ///  - If possible was modified, return true
    pub fn find_pairs(&mut self, cellgroup: &[usize]) -> (Vec<(Vec<usize>, Vec<u8>)>, bool) {
        // result[value] = indexes where it is located?
        // Step 1: find values, and their position(s).

        let initial = GenBits::new(self.width as usize);
        let mut value_positions: Vec<GenBits> = vec![initial; self.width as usize];

        // let cellgroup = self.group.group(Groups::Cell, cell_index);

        for value in 0..self.width {
            for pos in 0..self.width {
                let cellindex = cellgroup[pos as usize];

                if self.get(cellindex, value as usize + 1) {
                    // println!("Found {} ({},{}) = {}", cell_index, x, y, value);
                    value_positions[value as usize].set(pos as usize, true);
                }
            }
        }

        // Step 2: pair them up, and if complete, fixup possible (if needed).

        let width = self.width;
        let mut possibles_updated: bool = false;
        let mut pairs: Vec<(Vec<usize>, Vec<u8>)> = vec![];

        'outer: for idx in 0..width - 1 {
            // pos_found contains the positions.
            let pos_found = &value_positions[idx as usize];
            if pos_found.count_set() == 0 {
                continue;
            }
            if idx > 0 {
                // Check previous for matches - already checked...
                for check in 0..idx {
                    if *pos_found == value_positions[check as usize] {
                        continue 'outer;
                    }
                }
            }

            let count = pos_found.count_set();
            if count == 1 {
                // Single possible item here ... skip this for now.
                continue;
            }

            let mut matches = 1;
            // Pos contains the numbers in the cell.
            let mut pos = Vec::<u8>::new();
            pos.push(idx);

            if count != 1 {
                for look in idx + 1..width {
                    if value_positions[look as usize] == *pos_found {
                        matches += 1;
                        pos.push(look);
                    }
                }
            }

            if matches == count {
                // Ok! We found a pair (or triple, or quad, or ...)
                // Build new possible
                let mut new_possible = GenBits::new(width as usize);
                for p in &pos {
                    new_possible.set(*p as usize, true);
                }
                let mut index_pos = Vec::<usize>::new();

                for p in pos_found.iter() {
                    index_pos.push(p);
                    if self.possible[cellgroup[p as usize]] != new_possible {
                        // println!("Update pos {} with {:?}", p, new_possible);
                        /*
                        println!(
                            "Index is {} => {:?}",
                            cellgroup[p as usize],
                            self.group.xy(cellgroup[p as usize])
                        );
                        */
                        possibles_updated = true;
                        self.possible[cellgroup[p as usize]] = new_possible.clone();
                        // println!("Updated!");
                    }
                }
                pairs.push((index_pos, pos));
            }
        }

        (pairs, possibles_updated)
    }

    // This is working, it was code elsewhere that was having issues.

    #[must_use]
    /// Find positions where each value is possible.
    /// - Vec index is the value - 1  !
    /// - GenBits are the positions found.
    pub fn find_value_pos(&self) -> Vec<GenBits> {
        let zero = GenBits::new(self.max_index as usize);
        let mut result = vec![zero; self.width as usize];

        for index in 0..self.max_index {
            if self.possible[index as usize].count_set() == 0 {
                continue;
            }

            // Ok, there's some value there.
            for v in self.possible[index as usize].iter() {
                result[v as usize].set(index as usize, true);
            }
        }
        result
    }

    /// Display Possible
    /// - Collect all of the possibles, with max length.
    /// - Display formatted (using max length).
    pub fn display(&self) {
        let pos_info = self.pos_max();

        for y in 0..self.width {
            for x in 0..self.width {
                let idx = self.pos(x, y);

                print!("({},{}):{:3$} ", x + 1, y + 1, pos_info.0[idx], pos_info.1);
            }
            println!("");
        }
    }
}

impl Index<usize> for AnyPossible {
    type Output = GenBits;

    fn index<'a>(&'a self, i: usize) -> &'a GenBits {
        &self.possible[i]
    }
}

impl Index<(u8, u8)> for AnyPossible {
    type Output = GenBits;

    fn index<'a>(&'a self, i: (u8, u8)) -> &'a GenBits {
        &self.possible[self.pos(i.0, i.1)]
    }
}

impl IndexMut<usize> for AnyPossible {
    fn index_mut<'a>(&'a mut self, i: usize) -> &'a mut GenBits {
        &mut self.possible[i]
    }
}

impl IndexMut<(u8, u8)> for AnyPossible {
    fn index_mut<'a>(&'a mut self, i: (u8, u8)) -> &'a mut GenBits {
        let idx = self.pos(i.0, i.1);
        &mut self.possible[idx]
    }
}

/*
An idea I have for AnySolver is to store the AnyPossible as
sections.  row/column/cell.
I think this would better allow me to judge how hard the puzzle
is.  If it can be solved by looking at just one section --
that would be one level.  If it takes 2, another level. All 3,
yet another.  I think that's how I mentally judge the puzzles.

"Not all hard puzzles are hard."

 */

/*
pub struct Pairs {
    pub values: [u8; 2],
    pub pos: [u8; 2],
}
*/

#[derive(Debug, Clone)]
pub struct AnySolver {
    pub board: AnyBoard,
    pub possible: AnyPossible,
    pub group: AnyGroup,
    // Is this value set in the given cell?
    pub cell_poss: Vec<GenBits>,
}

impl AnySolver {
    /// Construct new solver from given board size. (3,4,5)
    pub fn new(board_size: u8) -> Self {
        let width: u8 = board_size * board_size;
        let mut default_poss = GenBits::new(width as usize);

        default_poss.set_bits(0..width as usize);

        let mut s = Self {
            board: AnyBoard::new(board_size),
            possible: AnyPossible::new(board_size),
            group: AnyGroup::new(board_size),
            cell_poss: vec![default_poss; width as usize * width as usize],
        };
        s.reset_possible();
        s
    }

    /// Construct new solver from given board.
    pub fn new_from(initial_board: &AnyBoard) -> Self {
        let mut default_poss = GenBits::new(initial_board.width as usize);
        default_poss.set_bits(0..initial_board.width as usize);

        let mut s = AnySolver {
            board: initial_board.clone(),
            possible: AnyPossible::new(initial_board.size),
            group: AnyGroup::new(initial_board.size),
            cell_poss: vec![
                default_poss;
                initial_board.width as usize * initial_board.width as usize
            ],
        };
        s.reset_possible();
        s
    }

    /// Clear out board and possible.
    /// - group is ok, unless they change the board size.
    pub fn clear(&mut self) {
        let board_size: u8 = self.board.size;
        self.board = AnyBoard::new(board_size);
        self.possible = AnyPossible::new(board_size);
        self.reset_possible();
    }

    /// Process a move
    /// - Remove value from rows, columns, and cells.
    /// - Clear possibles from (x,y) position, it's filled.
    /// - Added finalize parameter
    ///   - Use it when doing a single move.
    ///   - Don't do it when setting up the board.  Call finalize_move then.
    pub fn process_move(&mut self, x: u8, y: u8, value: u8, finalize: bool) {
        debug_assert!(
            x <= self.board.width && y <= self.board.width,
            "Expected ({}, {}) <= {}",
            x,
            y,
            self.board.width
        );
        debug_assert!(
            value <= self.board.width,
            "Expected value {} < {}",
            value,
            self.board.width
        );

        let mut g = self.group.row(y);
        let val: usize = value as usize;
        for g in g {
            self.possible.set(*g, val, false);
        }
        g = self.group.column(x);
        for g in g {
            self.possible.set(*g, val, false);
        }
        g = self.group.cell(self.group.which_cell(x, y));
        for g in g {
            self.possible.set(*g, val, false);
        }
        let idx = self.possible.pos(x, y);
        self.possible.possible[idx] = GenBits::new(self.board.width as usize);
        self.board.board[idx] = value;

        //
        // When working with GenBits, remember to use value -1
        //

        /*
        let ok = self.cell_poss[cell_index as usize].get(value as usize-1);
        if !ok {
            println!("{:?}", self.cell_poss);
            println!("Cell {} ({},{})={}", cell_index, x, y, value);
        }
        */

        // Tests are run as debug, and this kills things...
        // Like the invalid board in validated_board test.
        let cell_index = self.group.which_cell(x, y);

        /*
        Never use this code.  It asserts/panics on invalid boards.
        (Which is what validate_board checks...)

        debug_assert!(
            self.cell_poss[cell_index as usize].get(value as usize - 1),
            "Cell {} ({},{})={}",
            cell_index,
            x,
            y,
            value
        );
        */
        self.cell_poss[cell_index as usize].set(value as usize - 1, false);

        if finalize {
            self.finalize_possible();
            // self.finalize_cell(cell_index);
        }
        // Should I do the checks here for the logic possible fix?
        // Or, should I make it a different function?
        // Put somewhere else, since it has to re-check the entire board!

        // OR, just check row and columns?
    }

    /*
    This is very heavy to call (especially for 5X5 boards).
    Maybe call this when we're getting stuck? (before logic 2 pass?)

    Still to do: columns.  :P
     */

    pub fn finalize_possible(&mut self) -> bool {
        let width = self.board.width;
        let size = self.board.size;
        let mut updated: bool = false;
        let mut updated_here: bool = false;

        const OUTPUT: bool = false;

        println!("finalize_possible starts");

        // I might want to save the pair information.

        for cell in 0..width {
            let r = self
                .possible
                .find_pairs(self.group.group(Groups::Cell, cell));
            if r.1 {
                updated = true;
                updated_here = true;
            }
        }

        if updated_here {
            println!("find_pairs (update)");
        }
        updated_here = false;

        /// Do we output things from this function? (Debug, Development infomation)
        const NOISY: bool = true; // false;
        if OUTPUT {
            self.possible.display();
        }
        let mut cell_has_value = GenBits::new(self.board.max_index);
        // vec![false; self.board.max_index];

        /*
        let mut cell_update = vec![false; width as usize];
        cell_update[cell_index as usize] = true;
        */

        let values = self.possible.find_value_pos();

        // Not needed, displayed below (and only if not empty). ;)
        // println!("Values: {:?}", values);

        // Process ROWS, repeat for columns...

        let mut found_pos = Vec::<usize>::new();
        let mut cells_has = GenBits::new(self.board.width as usize);

        for row_index in 0..width {
            let row_indexes = self.group.row(row_index);
            // for row in 0..width {

            'outer: for value in 0..width {
                found_pos.clear();
                cells_has.clear();

                // Which indexes in this row contain this value?
                let g = &values[value as usize];

                let mut last_cell_found: u8 = 0;
                let mut row_found: usize = 0;
                let mut previously_found: bool = false;

                // See if row_indexes is contained in this.
                for ri in row_indexes.iter() {
                    if let Some(index) = g.iter().find(|x| *x == *ri) {
                        found_pos.push(index); // *ri);
                        let cell = self.group.which_cell_idx(index);

                        if previously_found {
                            // Check to see if it's different from the last one or not
                            if last_cell_found != cell {
                                // It is, stop looking... check next value
                                continue 'outer;
                            }
                        } else {
                            last_cell_found = cell;
                            // Which "row" of this cell?
                            row_found = *ri;
                            previously_found = true;
                        }

                        cells_has[self.group.which_cell_idx(index) as usize] = true;
                    }
                }

                if found_pos.len() == 0 {
                    continue;
                }

                // not needed now...
                if cells_has.count_set() > 1 {
                    continue;
                }

                if OUTPUT {
                    println!("cells_has: {} [see below]", cells_has);
                    self.possible.display();
                }

                /*
                Ok, I need to remove "value" from the other "rows" in this cell!

                This is working.

                Possible optimization idea here:  Save this info (if there's an update),
                so I don't try it again.  Short-circuit if I seen this on again.
                (It's already been done, so don't need to try it yet again...)

                This is a silly optimization.  The majority of the time, we're grinding
                over large groups of things that won't make any difference here.
                (The 5X5 solver in tests now takes ~70 seconds.)

                Perhaps this should be only called by the logic solver when it's stuck?!
                AND, this only is doing rows, we haven't done columns yet. :P

                 */

                let cell_start = self.group.cell_start(last_cell_found);
                let found_y = self.board.xy(row_found).1;
                if OUTPUT {
                    println!(
                        "Start {} ({},{}), found on {}",
                        last_cell_found, cell_start.0, cell_start.1, found_y
                    );
                }

                for update_row in 0..size {
                    let urow = cell_start.1 + update_row;
                    if urow == found_y {
                        continue;
                    }

                    for col in 0..size {
                        // let target = (cell_start.0 + col, urow);
                        // println!("Target: ({},{})", target.0, target.1);
                        if self.possible.set(
                            self.board.pos(cell_start.0 + col, urow),
                            value as usize + 1,
                            false,
                        ) {
                            updated = true;
                            updated_here = true;
                            // println!("Yes! to above target");
                        }
                    }
                }

                if updated_here {
                    if OUTPUT {
                        println!("Possible updated!  See!");
                        self.possible.display();
                    }
                }

                updated_here = false;

                /*
                if cells_has.count_set() > 1 {
                    continue;
                }
                */

                // Ok, here's the output of interest: (logic_test)
                // In row 1, the value 2 was found [6, 7, 8]
                // Or [(7, 1), (8, 1), (9, 1)]
                // All of these are within the same "cell"!
                // Maybe "short circuit" out early if they're in more then one cell?!?
                /*
                Index 6 (6,0) = cell 2
                Index 7 (7,0) = cell 2
                Index 8 (8,0) = cell 2
                cells_has: (2)
                In row 1, the value 2 was found [6, 7, 8]
                */
                // This is the spot where I want to be,
                // I'm not sure about the code below that outputs values...

                if OUTPUT {
                    println!(
                        "In row {}, the value {} was found {:?}",
                        row_index + 1,
                        value + 1,
                        found_pos
                    );
                    println!(
                        "Or {:?}",
                        found_pos
                            .iter()
                            .map(|x| self.board.xy(*x))
                            .map(|x| (x.0 + 1, x.1 + 1))
                            .collect::<Vec<_>>()
                    );
                }
            }

            // }
        }

        // Process COLUMNS

        for col_index in 0..width {
            let col_indexes = self.group.column(col_index);

            'outer: for value in 0..width {
                found_pos.clear();
                cells_has.clear();

                let g = &values[value as usize];
                let mut last_cell_found: u8 = 0;
                let mut col_found: usize = 0;
                let mut previously_found: bool = false;

                for ci in col_indexes.iter() {
                    if let Some(index) = g.iter().find(|x| *x == *ci) {
                        found_pos.push(index);
                        let cell = self.group.which_cell_idx(index);

                        if previously_found {
                            if last_cell_found != cell {
                                continue 'outer;
                            }
                        } else {
                            last_cell_found = cell;
                            col_found = *ci;
                            previously_found = true;
                        }

                        cells_has[self.group.which_cell_idx(index) as usize] = true;
                    }
                }

                if found_pos.len() == 0 {
                    continue;
                }
                // not needed
                if cells_has.count_set() > 1 {
                    continue;
                }

                if OUTPUT {
                    println!("col cells_has: {} [see below]", cells_has);
                    self.possible.display();
                }


                let cell_start = self.group.cell_start(last_cell_found);
                let found_x = self.board.xy(col_found).0;
                if OUTPUT {
                    println!(
                        "Start {} ({},{}), found on col {}",
                        last_cell_found, cell_start.0, cell_start.1, found_x
                    );
                }

                for update_col in 0..size {
                    let ucol = cell_start.0 + update_col;
                    if ucol == found_x {
                        continue;
                    }

                    for row in 0..size {
                        // let target = (cell_start.0 + col, urow);
                        // println!("Target: ({},{})", target.0, target.1);
                        if self.possible.set(
                            self.board.pos(ucol, cell_start.1 + row),
                            value as usize + 1,
                            false,
                        ) {
                            updated = true;
                            updated_here = true;
                            // println!("Yes! to above target");
                        }
                    }
                }

                if updated_here {
                    if OUTPUT {
                        println!("Possible updated (col)!  See!");
                        self.possible.display();
                    }
                }

            }
        }

        println!("finalize_possible ends...");
        return updated;

        /*
        for (value, positions) in values.iter().enumerate() {
            if positions.count_set() == 0 {
                // No positions, skip it.
                continue;
            }

            // reset
            cell_has_value.fill(false);

            for p in positions.iter() {
                cell_has_value[p] = true;
                // .set(p, true); // [p as usize] = true;
            }

            if NOISY {
                println!("Value {} : Pos {}", value, cell_has_value);
            }

            // cell_has_value gives me the indexes that contain value.
            // I need to process cell rows/columns...
            // Start with processing one (cell row), and refactor. (like always)

            // cells on left (because cells have the constraints).
            for cell in (0..size).map(|x| x * size) {
                // let c = self.group.cell(cell);
                let mut c = Vec::<&[usize]>::new();
                for row in 0..size {
                    c.push(self.group.cell(cell + row));
                }

                // rows go the entire width of the game...
                for _row in 0..width {}
            }
        }
        */

        /*
        let _pairs = self.possible
            .find_pairs(self.group.group(Groups::Cell, cell_index));
        */

        // println!("Finalize Cell Pairs: {:?}", self.possible.find_pairs(self.group.group(Groups::Cell, index)));

        // get positions of possible numbers - for every cell ?

        /*
        Great, I put my comments in a deprecated function...

        Let me describe what I want to do here:

        ╔═══╦═══╦═══╗
        ║ 2 ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║3  ║   ║   ║
        ║4  ║   ║   ║
        ║5  ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,3,4,5,6,7,8,9   (3,2):1,3,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,3,4,5,6,7,8,9   (3,3):1,3,4,5,6,7,8,9
        (1,4):                  (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9
        (1,6):                  (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):1,3,4,5,6,7,8,9   (3,7):1,2,3,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):1,3,4,5,6,7,8,9   (3,8):1,2,3,4,5,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,3,4,5,6,7,8,9   (3,9):1,2,3,4,5,6,7,8,9

        In cell 3, 2 must be in the last column.
        In cell 6, 2 must be in the first column.  (3,7), (3,8) and (3,9) should not contain 2.

        Top cell, 2 is in the middle (2,_). (it's not in the list of possible.)
        Middle cell, 2 can only be in the last column. (3,*)
        Bottom cell, 2 must be in (1,*).

        How should I determine that?

        0.)
        Build a table of numbers, and their possible positions. (used in the next steps)
        I probably need some helpers to give me a row/column of a given position... as well.
        row/column iterators ?

        1.)
        In a given cell's column/row, if a number is possible in that cell's column/row, remove it from the other cell's row/column.
        (Must be there, so remove it.)
        (3,7), (3,8), (3,9) remove 2.

        2.)
        In a given cell's column/row, if a number isn't possible in the same column/row in the other cells
          -- remove that number as possible from that cell's other row/columns.

        When updating the possibles - track which cells changed (so we can check/recheck just those).
        After finalize, clear the changed flag (for that cell).

        When allocating the blocks, possibly use the max value (size) -- this should keep the allocated blocks
        the same (hopefully keep reusing the same size block of memory over and over)
        -- or allocate it once and just reuse/init the same block over and over ? (reduce allocation/deallocations)

        Also the reverse should be possible:
        ╔═══╦═══╦═══╗
        ║T  ║   ║   ║
        ║T  ║   ║   ║
        ║T  ║   ║   ║
        ╠═══╬═══╬═══╣
        ║3tt║   ║   ║
        ║4tt║   ║   ║
        ║5tt║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║2  ║   ║
        ║   ║   ║2  ║
        ║7tt║   ║   ║
        ╚═══╩═══╩═══╝

        2's possible in column 2 and 3. (for cells 3 and 6).  Represented by 't'.
        So, 2 must be in column 1 of cell 0.  Represented by 'T'.


         */

        println!("finalize_possible ends...");
        updated
    }

    // I'm saving finalize_cell, just because it's working,
    // but has a terrible name.  :(
    #[deprecated]
    pub fn finalize_cell(&mut self, cell_index: u8) {
        let width = self.board.width;
        let size = self.board.size;
        let mut updated: bool = false;

        // Ok, right now this slows things down immensely, (tests-solver)
        // with the output, and calculations...

        // return;

        // let mut update: bool = false;
        /// Do we output things from this function? (Debug, Development infomation)
        const NOISY: bool = true; // false;
                                  // use width, not size. ;)
        self.possible.display();
        let mut cell_has_value = GenBits::new(self.board.max_index);
        // vec![false; self.board.max_index];

        let mut cell_update = vec![false; width as usize];
        cell_update[cell_index as usize] = true;

        let values = self.possible.find_value_pos();

        // Not needed, displayed below (and only if not empty). ;)
        // println!("Values: {:?}", values);

        for row_index in 0..width {
            let row_indexes = self.group.row(row_index);
            // for row in 0..width {
            let mut found_pos = Vec::<usize>::new();
            let mut cells_has = GenBits::new(self.board.width as usize);

            'outer: for value in 0..width {
                found_pos.clear();
                cells_has.clear();

                // Which indexes in this row contain this value?
                let g = &values[value as usize];

                let mut last_cell_found: u8 = 0;
                let mut row_found: usize = 0;
                let mut previously_found: bool = false;

                // See if row_indexes is contained in this.
                for ri in row_indexes.iter() {
                    if let Some(index) = g.iter().find(|x| *x == *ri) {
                        found_pos.push(index); // *ri);
                        let cell = self.group.which_cell_idx(index);

                        if previously_found {
                            // Check to see if it's different from the last one or not
                            if last_cell_found != cell {
                                // It is, stop looking... check next value
                                continue 'outer;
                            }
                        } else {
                            last_cell_found = cell;
                            // Which "row" of this cell?
                            row_found = *ri;
                            previously_found = true;
                        }

                        cells_has[self.group.which_cell_idx(index) as usize] = true;
                    }
                }

                if found_pos.len() == 0 {
                    continue;
                }

                // not needed now...
                if cells_has.count_set() > 1 {
                    continue;
                }

                println!("cells_has: {} [see below]", cells_has);
                self.possible.display();

                /*
                Ok, I need to remove "value" from the other "rows" in this cell!

                This is working.

                Possible optimization idea here:  Save this info (if there's an update),
                so I don't try it again.  Short-circuit if I seen this on again.
                (It's already been done, so don't need to try it yet again...)
                 */

                let cell_start = self.group.cell_start(last_cell_found);
                let found_y = self.board.xy(row_found).1;
                println!(
                    "Start {} ({},{}), found on {}",
                    last_cell_found, cell_start.0, cell_start.1, found_y
                );

                for update_row in 0..size {
                    let urow = cell_start.1 + update_row;
                    if urow == found_y {
                        continue;
                    }

                    for col in 0..size {
                        let target = (cell_start.0 + col, urow);
                        // println!("Target: ({},{})", target.0, target.1);
                        if self.possible.set(
                            self.board.pos(cell_start.0 + col, urow),
                            value as usize + 1,
                            false,
                        ) {
                            updated = true;
                            // println!("Yes! to above target");
                        }
                    }
                }

                if updated {
                    println!("Possible updated!  See!");
                    self.possible.display();
                }
                /*
                if cells_has.count_set() > 1 {
                    continue;
                }
                */

                // Ok, here's the output of interest: (logic_test)
                // In row 1, the value 2 was found [6, 7, 8]
                // Or [(7, 1), (8, 1), (9, 1)]
                // All of these are within the same "cell"!
                // Maybe "short circuit" out early if they're in more then one cell?!?
                /*
                Index 6 (6,0) = cell 2
                Index 7 (7,0) = cell 2
                Index 8 (8,0) = cell 2
                cells_has: (2)
                In row 1, the value 2 was found [6, 7, 8]
                */
                // This is the spot where I want to be,
                // I'm not sure about the code below that outputs values...

                println!(
                    "In row {}, the value {} was found {:?}",
                    row_index + 1,
                    value + 1,
                    found_pos
                );
                println!(
                    "Or {:?}",
                    found_pos
                        .iter()
                        .map(|x| self.board.xy(*x))
                        .map(|x| (x.0 + 1, x.1 + 1))
                        .collect::<Vec<_>>()
                );
            }

            // }
        }

        return;

        for (value, positions) in values.iter().enumerate() {
            if positions.count_set() == 0 {
                // No positions, skip it.
                continue;
            }

            // reset
            cell_has_value.fill(false);

            for p in positions.iter() {
                cell_has_value[p] = true;
                // .set(p, true); // [p as usize] = true;
            }

            if NOISY {
                println!("Value {} : Pos {}", value, cell_has_value);
            }

            // cell_has_value gives me the indexes that contain value.
            // I need to process cell rows/columns...
            // Start with processing one (cell row), and refactor. (like always)

            // cells on left (because cells have the constraints).
            for cell in (0..size).map(|x| x * size) {
                // let c = self.group.cell(cell);
                let mut c = Vec::<&[usize]>::new();
                for row in 0..size {
                    c.push(self.group.cell(cell + row));
                }

                // rows go the entire width of the game...
                for _row in 0..width {}
            }
        }
        /*
        let _pairs = self.possible
            .find_pairs(self.group.group(Groups::Cell, cell_index));
        */

        // println!("Finalize Cell Pairs: {:?}", self.possible.find_pairs(self.group.group(Groups::Cell, index)));

        // get positions of possible numbers - for every cell ?

        /*
        Great, I put my comments in a deprecated function...

        Let me describe what I want to do here:

        ╔═══╦═══╦═══╗
        ║ 2 ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║3  ║   ║   ║
        ║4  ║   ║   ║
        ║5  ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,3,4,5,6,7,8,9   (3,2):1,3,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,3,4,5,6,7,8,9   (3,3):1,3,4,5,6,7,8,9
        (1,4):                  (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9
        (1,6):                  (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):1,3,4,5,6,7,8,9   (3,7):1,2,3,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):1,3,4,5,6,7,8,9   (3,8):1,2,3,4,5,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,3,4,5,6,7,8,9   (3,9):1,2,3,4,5,6,7,8,9

        In cell 3, 2 must be in the last column.
        In cell 6, 2 must be in the first column.  (3,7), (3,8) and (3,9) should not contain 2.

        Top cell, 2 is in the middle (2,_). (it's not in the list of possible.)
        Middle cell, 2 can only be in the last column. (3,*)
        Bottom cell, 2 must be in (1,*).

        How should I determine that?

        0.)
        Build a table of numbers, and their possible positions. (used in the next steps)
        I probably need some helpers to give me a row/column of a given position... as well.
        row/column iterators ?

        1.)
        In a given cell's column/row, if a number is possible in that cell's column/row, remove it from the other cell's row/column.
        (Must be there, so remove it.)
        (3,7), (3,8), (3,9) remove 2.

        2.)
        In a given cell's column/row, if a number isn't possible in the same column/row in the other cells
          -- remove that number as possible from that cell's other row/columns.

        When updating the possibles - track which cells changed (so we can check/recheck just those).
        After finalize, clear the changed flag (for that cell).

        When allocating the blocks, possibly use the max value (size) -- this should keep the allocated blocks
        the same (hopefully keep reusing the same size block of memory over and over)
        -- or allocate it once and just reuse/init the same block over and over ? (reduce allocation/deallocations)

        Also the reverse should be possible:
        ╔═══╦═══╦═══╗
        ║T  ║   ║   ║
        ║T  ║   ║   ║
        ║T  ║   ║   ║
        ╠═══╬═══╬═══╣
        ║3tt║   ║   ║
        ║4tt║   ║   ║
        ║5tt║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║2  ║   ║
        ║   ║   ║2  ║
        ║7tt║   ║   ║
        ╚═══╩═══╩═══╝

        2's possible in column 2 and 3. (for cells 3 and 6).  Represented by 't'.
        So, 2 must be in column 1 of cell 0.  Represented by 'T'.


         */
        let (cell_x, cell_y) = self.group.xy(self.group.group(Groups::Cell, cell_index)[0]);
        let x_range = cell_x..cell_x + self.board.size;
        let y_range = cell_y..cell_y + self.board.size;
        let top_cell = cell_index - (cell_index / size) * size;
        let left_cell = cell_index - (cell_index % size);

        if NOISY {
            println!(
                "finalize_cell({cell_index}): ({cell_x},{cell_y}) top {top_cell} left {left_cell}"
            );
        }

        // Ranges are consumed.

        // This looks like it does what I'm doing somewhere else!
        // TODO:
        // TOFIX:
        // See: AnyPossible's find_value_pos!

        // Check columns - (of given cell) for logical removes (see logic_test for example)
        for value in 0..width {
            // Find what columns value is allowed. (in which cells).
            // Is this value needed in any of the cells?
            let mut found = GenBits::new(width as usize);
            let mut all_cells = found.clone();

            for ci in (top_cell..width).step_by(size as usize) {
                all_cells.set(ci as usize, true);
                if self.cell_poss[ci as usize].get(value as usize) {
                    found.set(ci as usize, true);
                }
            }

            if found.count_set() == 0 {
                // Not possible anywhere, so skip to next value.
                continue;
            }

            // println!("Value {} {:?} is needed {:?}", value + 1, all_cells, found);

            for _x in x_range.clone() {
                for _y in 0..width {
                    // Calculate the row and column for the given cell.
                    // let _col = (cell_index % width) * width + c;
                    // println!("Index {} COL {}", index, col);
                }
            }
        }

        // Check rows
        for _value in 0..width {
            for _y in y_range.clone() {
                for _x in 0..width {
                    // Calculate the row and column for the given cell.
                    // let _row = (cell_index / width) * width + r;
                    // println!("Index {} ROW {}", index, row);
                }
            }
        }
        if NOISY {
            println!("finalize_cell ends...");
        }
    }

    #[deprecated]
    /// Finalize move(s) on the board.
    /// - This checks the board for logic fixes. (when column is full/has 1+pair/triple)
    pub fn finalize_move(&mut self) {
        // Process the cells diagonally.
        println!("TODO: (naively)");

        // naive solution - check all the cells.
        self.finalize_possible();

        /*
        for idx in 0..self.board.width {
            self.finalize_cell(idx);
        }
        */

        /*
        for idx in 0..self.board.size {
            self.finalize_cell(idx + idx * self.board.size);
        }
        */
        println!("TODO");
    }

    /// Validate the board
    /// Reuse reset_possible code, verify the values are possible.
    /// - This does not check if the board is completed.
    /// - It does check that blanks have possible values.
    pub fn validate_board(&mut self) -> bool {
        let mut has_blanks = false;

        self.possible.clear();
        // reset cell possible
        let mut default_poss = GenBits::new(self.board.width as usize);
        default_poss.set_bits(0..self.board.width as usize);

        for x in 0..self.board.width {
            self.cell_poss[x as usize] = default_poss.clone();
        }

        for y in 0..self.board.width {
            for x in 0..self.board.width {
                let value = self.board.get(x, y);

                if value != 0 {
                    if !self.possible.get(self.possible.pos(x, y), value as usize) {
                        // I was going to reset_possible, but the board is broken!
                        // Leave in a broken state.

                        // log (maybe)
                        // println!("Invalid at ({},{}) can't place {}.", x + 1, y + 1, value);
                        // self.board.display();
                        return false;
                    }
                    self.process_move(x, y, value, false);
                } else {
                    has_blanks = true;
                }
            }
        }
        // Ok, the pieces given fit correctly with the sudoku constraints.
        self.finalize_possible();
        // self.finalize_move();

        if has_blanks {
            // Verify that the remaining value == 0 positions have possibles.
            // - If they are blank, then it isn't a valid puzzle!
            for y in 0..self.board.width {
                for x in 0..self.board.width {
                    let value = self.board.get(x, y);
                    if value == 0 {
                        let count = self.possible.possible[self.possible.pos(x, y)].count_set();
                        if count == 0 {
                            // log (maybe)
                            // println!("Invalid ({},{}) = no values possible.", x + 1, y + 1);
                            // self.board.display();
                            return false;
                        }
                    }
                }
            }
        }
        true
    }

    /// Reset the possible
    /// - For when a new puzzle has been loaded.
    /// - When something has changed, and the possibles are out of sync.
    pub fn reset_possible(&mut self) {
        self.possible.clear();

        // reset cell possible
        let mut default_poss = GenBits::new(self.board.width as usize);
        default_poss.set_bits(0..self.board.width as usize);

        for x in 0..self.board.width {
            self.cell_poss[x as usize] = default_poss.clone();
        }

        for y in 0..self.board.width {
            for x in 0..self.board.width {
                let value = self.board.get(x, y);
                if value != 0 {
                    self.process_move(x, y, value, false);
                }
            }
        }
        self.finalize_possible();
        // self.finalize_move();
    }

    /// set (x,y) to value.
    /// - This updates the board.
    /// - This updates all the possibles (row,column,cell).
    /// - Clears the possible for (x,y) [See process_move].
    /// TO FIX:  I think set also needs finalize ... to pass to process_move.
    pub fn set(&mut self, x: u8, y: u8, value: u8, finalize: bool) {
        debug_assert!(
            x < self.board.width && y < self.board.width,
            "Expected ({}, {}) < {}",
            x,
            y,
            self.board.width
        );
        debug_assert!(
            value < self.board.width + 1,
            "Expected value {} < {}",
            value,
            self.board.width + 1
        );

        self.board.set(x, y, value);
        if value != 0 {
            self.process_move(x, y, value, finalize);
        }
    }

    #[must_use]
    pub fn get(&self, x: u8, y: u8) -> u8 {
        debug_assert!(
            x < self.board.width && y < self.board.width,
            "Expected ({}, {}) < {}",
            x,
            y,
            self.board.width
        );
        self.board.get(x, y)
    }

    /// Make completed board.
    /// - uses fill_board (recursive)
    pub fn make(&mut self) {
        let mut rng = ChaCha20Rng::from_entropy();
        self.reset_possible();
        self.fill_board(&mut rng);
    }

    /// Recursively completely fill the board.
    /// This takes - quite a bit of time - for a 25x25 board.
    fn fill_board(&mut self, rng: &mut ChaCha20Rng) -> bool {
        let backup = self.clone();
        let max_index = self.board.max_index;
        let width = self.board.width;

        for idx in 0..max_index {
            if self.board.board[idx] == 0 {
                let (x, y) = self.board.xy(idx);
                let mut available: Vec<u8> = Vec::new();
                available.reserve(width as usize);

                for t in self.possible.possible[idx].iter() {
                    available.push((t + 1) as u8);
                }
                if available.len() == 0 {
                    return false;
                }

                // print!("{:?}", available);
                // Randomize the possible items.
                available.shuffle(rng);
                // available.as_mut_slice().shuffle(rng);

                // print!("{:?}", available);
                for value in available.into_iter() {
                    assert!(value != 0);

                    self.set(x, y, value, true);
                    // self.board.display();
                    // self.possible.display();

                    if self.fill_board(rng) {
                        return true;
                    }
                    // Failure
                    self.board.copy(&backup.board);
                    self.possible.copy(&backup.possible);

                    for i in 0..backup.cell_poss.len() {
                        self.cell_poss[i] = backup.cell_poss[i].clone();
                    }
                    // self.cell_poss.copy( backup.cell_poss);
                }

                // We've run out of possible.
                return false;
            }
        }
        // We've visited everything, and it isn't 0.
        true
    }

    /// Make a puzzle by removing items
    /// - Verify it still has a single solution.
    pub fn make_puzzle(&mut self) {
        let mut rng = ChaCha20Rng::from_entropy();

        if !self.board.complete() {
            self.fill_board(&mut rng);
        }
        self.reset_possible();
        while self.remove(&mut rng) {}
    }

    /// Remove an item
    /// - Remove a number, test if still solvable.
    /// - Returns true when removal was successful, and still solvable.
    /// - Return false when it can't.
    fn remove(&mut self, rng: &mut ChaCha20Rng) -> bool {
        let puzrange = Uniform::new(0, self.board.width);
        let mut x = 0;
        let mut y = 0;
        let mut value: u8 = 0;

        while value == 0 {
            x = puzrange.sample(rng);
            y = puzrange.sample(rng);
            value = self.get(x, y);
        }

        self.set(x, y, 0, true);

        // clone, and solve by logic.
        let mut puzcopy = self.clone();
        puzcopy.reset_possible();
        puzcopy.solve_logic();

        if puzcopy.board.complete() {
            return true;
        }

        // Not solvable, restore and return false.
        // Since this failed, don't bother to finalize.
        self.set(x, y, value, false);
        return false;
    }

    fn logic_pass1(&mut self) -> bool {
        // Pass 1: Look for singles in the possible sets.
        let mut pass1 = false;

        // Repeat pass 1 until all have been found.
        let mut pass1_again = true;

        while pass1_again {
            // Pass 1 - look for singles.
            pass1_again = false;
            let mut found_something = false;

            for i in 0..self.possible.max_index {
                if self.board.board[i] != 0 {
                    // Skip, if position already filled.
                    continue;
                }

                if self.possible.possible[i].count_set() == 1 {
                    // Get value
                    let value = self.possible.possible[i].iter().next().unwrap() + 1;
                    let pos = self.board.xy(i);
                    // println!("SET {}({},{})={}", i, pos.0 + 1, pos.1 + 1, value);
                    self.set(pos.0, pos.1, value as u8, true);
                    found_something = true;
                }
            }

            if found_something {
                // We found something on this pass, so:
                // - Try it again.
                // - Record that we did something.
                pass1 = true;
                pass1_again = true;
            }
        }
        pass1
    }

    fn logic_pass2(&mut self) -> bool {
        let mut found_something = false;

        let mut pass2 = false;
        let mut pass2_again = true;

        let width = self.group.width;
        let grp = self.group.clone();

        while pass2_again {
            pass2_again = false;

            // let mut values = Bits(0); // HashSet<u8> = HashSet::new();
            let mut values = GenBits::new(self.board.width as usize);

            let mut group_process = |this: &mut Self, grp: &[usize]| {
                // Collect all the possible values within the group.
                values.clear();
                for gidx in 0..width {
                    // println!("possible: {:?}", this.possible[grp.items[gidx as usize] as usize]);
                    for v in this.possible.possible[grp[gidx as usize]].iter() {
                        values.set(v as usize, true);
                    }
                    // values.extend(this.possible[grp.0[gidx as usize] as usize]);
                    // println!("now     : {:?}", this.possible[grp.items[gidx as usize] as usize]);
                }

                // println!("values {:?}", values);

                // Now, check for singles.
                for v in values.iter() {
                    // println!("Check Value: {}", v);

                    let mut count = 0;
                    let mut pos = 0;

                    for gidx in 0..width {
                        if this.possible.possible[grp[gidx as usize]].get(v as usize) {
                            // if this.possible[grp.0[gidx as usize] as usize].contains(&v) {
                            count += 1;
                            pos = grp[gidx as usize];
                            if count > 1 {
                                break;
                            } else {
                                // print!(" IDX {} POS {} ", gidx, pos);
                            }
                        }
                    }
                    if count == 1 {
                        // don't need this, it was v!
                        // let value = this.possible[pos as usize].iter().next().cloned().unwrap();
                        let xy = this.board.xy(pos);
                        // this.possible.display();
                        // this.board.display();
                        this.set(xy.0, xy.1, (v + 1) as u8, true);
                        // println!("SET {} ({},{}) = {}", pos, xy.0 + 1, xy.1 + 1, v+1);

                        found_something = true;
                    }
                }
            };

            for gr in Groups::iter() {
                for i in 0..width {
                    let g = grp.group(gr, i);
                    group_process(self, g);
                }
            }

            if found_something == true {
                // Ok, pass 2 found something.
                pass2 = true;
                found_something = false;
                pass2_again = true;
                // continue;
            }
        }
        pass2
    }

    /// Solve by logic alone.
    /// - Returns true if solved.
    /// - It might not be solved (if guessing is required).
    pub fn solve_logic(&mut self) -> bool {
        // self.reset_possible(); // destroys anything pass3 accomplishes ...

        // self.board.display();
        // self.possible.display();

        let mut found_more = true;

        while found_more {
            found_more = false;

            let mut result1 = true;
            // Pass 1: Look for singles in the possible sets.
            while result1 {
                result1 = self.logic_pass1();
                /*
                if result1 {
                    println!("Pass1");
                };
                */
            }

            let mut result2 = true;
            while result2 {
                result2 = self.logic_pass2();
                if result2 {
                    // println!("Pass2");
                    found_more = true;
                };
            }

            /*
            // Pass 2: Is the same as Pass 1!
            // We just look in groups, instead of all the indexes.

            let mut found_something = false;
            let mut pass1 = false;

            // Repeat pass 1 until all have been found.
            let mut pass1_again = true;

            let width = self.group.width;
            let grp = self.group.clone();
            let mut pass2 = false;

            while pass1_again {
                // Pass 1 - look for singles.

                for i in 0..self.possible.max_index {
                    if self.board.board[i] != 0 {
                        // Skip, if position already filled.
                        continue;
                    }

                    if self.possible.possible[i].count_set() == 1 {
                        // Get value
                        let value = self.possible.possible[i].iter().next().unwrap() + 1;
                        let pos = self.board.xy(i);
                        // println!("SET {}({},{})={}", i, pos.0 + 1, pos.1 + 1, value);
                        self.set(pos.0, pos.1, value);
                        found_something = true;
                    }
                }

                if found_something {
                    // We found something this pass.
                    // - set pass1 to true (pass 1 found something)
                    // - reset found_something so we can try again.
                    pass1 = true;
                    found_something = false;
                    continue;
                } else {
                    // Nothing found with this run of pass 1.

                    // Pass 2 - Look for singles within the groups.

                    // Are we done?
                    /*
                    if self.board.complete() {
                        return false;
                    }
                    // We're getting stuck in pass 2 ...
                    // not anymore.  ;)
                    println!("Pass 2:");
                    self.board.display();

                    */

                    found_something = false;

                    // let mut values = Bits(0); // HashSet<u8> = HashSet::new();
                    let mut values = GenBits::<u32>(0);

                    let mut group_process = |this: &mut Self, grp: &[usize]| {
                        // Collect all the possible values within the group.
                        values.clear();
                        for gidx in 0..width {
                            // println!("possible: {:?}", this.possible[grp.items[gidx as usize] as usize]);
                            for v in this.possible.possible[grp[gidx as usize]].iter() {
                                values.set(v as usize, true);
                            }
                            // values.extend(this.possible[grp.0[gidx as usize] as usize]);
                            // println!("now     : {:?}", this.possible[grp.items[gidx as usize] as usize]);
                        }

                        // println!("values {:?}", values);

                        // Now, check for singles.
                        for v in values.iter() {
                            // println!("Check Value: {}", v);

                            let mut count = 0;
                            let mut pos = 0;

                            for gidx in 0..width {
                                if this.possible.possible[grp[gidx as usize]].get(v as usize) {
                                    // if this.possible[grp.0[gidx as usize] as usize].contains(&v) {
                                    count += 1;
                                    pos = grp[gidx as usize];
                                    if count > 1 {
                                        break;
                                    } else {
                                        // print!(" IDX {} POS {} ", gidx, pos);
                                    }
                                }
                            }
                            if count == 1 {
                                // don't need this, it was v!
                                // let value = this.possible[pos as usize].iter().next().cloned().unwrap();
                                let xy = this.board.xy(pos);
                                // this.possible.display();
                                // this.board.display();
                                this.set(xy.0, xy.1, v + 1);
                                // println!("SET {} ({},{}) = {}", pos, xy.0 + 1, xy.1 + 1, v+1);

                                found_something = true;
                            }
                        }
                    };

                    // Change to 0..WIDTH ...  Keep it simple.
                    for i in 0..width {
                        // println!("Column {i}:");
                        let mut g = grp.column(i);
                        group_process(self, g);
                        // println!("Row {i}:");
                        g = grp.row(i);
                        group_process(self, g);
                        // println!("Cell {i}:");
                        g = grp.cell(i);
                        group_process(self, g);
                    }

                    if found_something == true {
                        pass2 = true;
                        // Ok, pass 2 found something.
                        found_something = false;
                        continue;
                    }

                    //
                    // - If pass1 set, reset the found_something (because it
                    //   did find something)
                    // - Clear the pass1_again flag, we're done with pass 1.
                    if pass1 {
                        pass1_again = false;
                        found_something = true;
                    } else {
                        // Ok, we didn't find anything.
                        // Break out of loop, get unstuck.
                        break;
                    }
                }
            }

            // Pass 3:
            // - Find pairs & remove those numbers from corresponding group.

            let grp = self.group.clone();

            println!("Pass 3:");
            self.board.display();
            self.possible.display();

            assert_eq!(self.board.width, self.possible.width);

            // Pair processing.
            for i in 0..width {
                let mut g = grp.cell(i);
                println!("Cell {}: {:?}", i, g);

                for gidx in 0..WIDTH - 1 {
                    let gpos = g[gidx as usize];
                    if self.possible.possible[gpos].count_set() == 2 {
                        // Found a pair

                        for fidx in gidx + 1..width {
                            let fpos = g[fidx as usize];
                            if self.possible.possible[fpos as usize].count_set() == 2 {
                                // Ok, there's another pair
                                // if self.possible[gpos as usize].is_subset(&self.possible[fpos as usize])

                                if self.possible.possible[gpos] == self.possible.possible[fpos] {
                                    // Ok, they have the same values!
                                    // Ok, remove the items in the pair from the cell.
                                    // Don't touch the gpos/fpos records.  Keep those!

                                    // This looks .. WRONG!  only z+1 when using set(value)!

                                    let mut values: [u8; 2] = [0, 0];
                                    let mut vpos = 0;
                                    for z in self.possible.possible[gpos].iter() {
                                        values[vpos] = z;
                                        vpos += 1;
                                    }

                                    // Ok, other then being off by 1 (because x,y starts at 0 not 1),
                                    // This is, at least, displaying the information properly.

                                    self.possible.display();
                                    println!(
                                        "Pairs {gpos}({},{}) and {fpos}({},{}) {:?}",
                                        self.possible.xy(gpos).0 + 1,
                                        self.possible.xy(gpos).1 + 1,
                                        self.possible.xy(fpos).0 + 1,
                                        self.possible.xy(fpos).1 + 1,
                                        values
                                    );

                                    let mut pair_removed = false;

                                    // Check to see if anything was removed.

                                    for remove in 0..width {
                                        if (gidx == remove) || (fidx == remove) {
                                            // Skip the found pair indexes.  Don't remove those!
                                            continue;
                                        }

                                        // Ok, these aren't the ones to save, so:
                                        let rpos = g[remove as usize];
                                        if self.possible.possible[rpos].get(values[0] as usize) {
                                            self.possible.possible[rpos].set(values[0] as usize, false);
                                            found_something = true;
                                            pair_removed = true;
                                            println!("Removed {} from {}({},{})", values[0], rpos, self.possible.xy(rpos).0, self.possible.xy(rpos).1);
                                        }

                                        if self.possible.possible[rpos].get(values[1] as usize) {
                                            self.possible.possible[rpos].set(values[1] as usize, false);
                                            found_something = true;
                                            pair_removed = true;
                                            println!("Removed {} from {}({},{})", values[1], rpos, self.possible.xy(rpos).0, self.possible.xy(rpos).1);
                                        }
                                    }

                                    if pair_removed {
                                        println!("pair removed...");
                                        println!(
                                            "--> Pairs {gpos}({},{}) and {fpos}({},{}) {:?}",
                                            self.possible.xy(gpos).0 + 1,
                                            self.possible.xy(gpos).1 + 1,
                                            self.possible.xy(fpos).0 + 1,
                                            self.possible.xy(fpos).1 + 1,
                                            values
                                        );
                                    }

                                    // Check the x's and y's to see if we can also process a row/column too.
                                    if self.possible.xy(gpos).0 == self.possible.xy(fpos).0 {
                                        // Matching X - process column
                                        let column = xy(gpos).0;

                                        vpos = 0;
                                        for z in self.possible.possible[gpos].iter() {
                                            values[vpos] = z + 1;
                                            vpos += 1;
                                        }
                                        for remove in 0..WIDTH {
                                            if (remove == xy(gpos).1) || (remove == xy(fpos).1) {
                                                continue;
                                            }

                                            let pos = self.possible.pos(column, remove);

                                            if self.possible.possible[pos].get(values[0] as usize) {
                                                self.possible.possible[pos]
                                                    .set(values[0] as usize, false);
                                                found_something = true;
                                                pair_removed = true;
                                            }

                                            if self.possible.possible[pos].get(values[1] as usize) {
                                                self.possible.possible[pos]
                                                    .set(values[1] as usize, false);
                                                found_something = true;
                                                pair_removed = true;
                                            }
                                        }
                                    }

                                    if self.possible.xy(gpos).1 == self.possible.xy(fpos).1 {
                                        // Matching Y - process row
                                        let row = self.possible.xy(gpos).1;

                                        vpos = 0;
                                        for z in self.possible.possible[gpos].iter() {
                                            values[vpos] = z + 1;
                                            vpos += 1;
                                        }
                                        for remove in 0..width {
                                            if (remove == self.possible.xy(gpos).0)
                                                || (remove == self.possible.xy(fpos).0)
                                            {
                                                continue;
                                            }
                                            let pos = self.possible.pos(remove, row);

                                            if self.possible.possible[pos].get(values[0] as usize) {
                                                self.possible.possible[pos]
                                                    .set(values[0] as usize, false);
                                                found_something = true;
                                                pair_removed = true;
                                            }

                                            if self.possible.possible[pos].get(values[1] as usize) {
                                                self.possible.possible[pos]
                                                    .set(values[1] as usize, false);
                                                found_something = true;
                                                pair_removed = true;
                                            }
                                        }
                                    }

                                    if pair_removed {
                                        if true {
                                            println!(
                                                "Pair found! {} {}: {} {:?} and {} {:?} !",
                                                gidx,
                                                fidx,
                                                gpos,
                                                self.board.xy(gpos),
                                                fpos,
                                                self.board.xy(fpos)
                                            );
                                            self.possible.display();
                                            // panic!("... We're lost ...");
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            println!("Pass 3 - Ending...");
            found_something
            */
        }
        self.board.complete()
    }
}

#[cfg(test)]

mod tests {
    use crate::sudoku::*;

    #[test]
    /// Test indexing the board by []
    fn board_index_test() {
        let mut board = AnyBoard::new(3);
        board[(1, 0)] = 2;
        board[(0, 3)] = 3;
        board[(0, 4)] = 4;
        board[(0, 5)] = 5;
        board[3 + 6 * 3] = 8;
        board.display();

        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " 2       ",
                "         ",
                "   8     ",
                "3        ",
                "4        ",
                "5        ",
                "         ",
                "         ",
                "         "
            ]
        );
    }

    #[test]
    fn verify_validate_board_does_not_destroy_possible_test() {
        let mut board = AnyBoard::new(3);
        let result = board.load_from_tld(
            'b',
            '_',
            "__c_____e_h__cb___bd___ei_ch_jb__d___________i_eh__b__dg___ij_f_i__jg_____b_____g",
        );
        assert!(result.is_ok());
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "  17 83  ",
                " 73   68 ",
                "2  9 4  1",
                "   1 7   ",
                " 2     9 ",
                " 14   86 ",
                "  83 19  ",
                "         ",
                "4 2   5 6"
            ]
        );

        board.display();
        let mut solver = AnySolver::new_from(&board);
        let possible = solver.possible.clone();

        assert!(solver.validate_board());
        assert_eq!(possible.possible, solver.possible.possible);
        solver.reset_possible();

        /*
        if !solver.solve_logic() {
            println!("Failed to solve by logic!");
            solver.board.display();
        }
        solver.board.display();
        */

        /*
        ╔═══╦═══╦═══╗
        ║541║768║329║
        ║973║512║684║
        ║286║934║751║
        ╠═══╬═══╬═══╣
        ║869║157║243║
        ║325║486║197║
        ║714║293║865║
        ╠═══╬═══╬═══╣
        ║658║341║972║
        ║197║625║438║
        ║432║879║516║
        ╚═══╩═══╩═══╝
         */

        /*
        assert!(solver.validate_board());
        assert!(solver.board.complete());

        let strings = solver.board.to_strings();
        assert_eq!(
            strings,
            vec![
                "541768329",
                "973512684",
                "286934751",
                "869157243",
                "325486197",
                "714293865",
                "658341972",
                "197625438",
                "432879516"
            ]
        );
        */
    }

    #[ignore = "Finalize not completed (NNY!)"]
    #[test]
    fn logic_test() {
        let mut board = AnyBoard::new(3);
        board.set(0, 1, 2);
        board.set(3, 0, 3);
        board.set(4, 0, 4);
        board.set(5, 0, 5);
        board.display();
        /*
        ╔═══╦═══╦═══╗
        ║   ║345║   ║
        ║2  ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
         */
        let solver = AnySolver::new_from(&board);
        solver.possible.display();
        /*
        (1,1):1,6,7,8,9         (2,1):1,6,7,8,9         (3,1):1,6,7,8,9         (4,1):                  (5,1):                  (6,1):                  (7,1):1,2,6,7,8,9       (8,1):1,2,6,7,8,9       (9,1):1,2,6,7,8,9
        (1,2):                  (2,2):1,3,4,5,6,7,8,9   (3,2):1,3,4,5,6,7,8,9   (4,2):1,6,7,8,9         (5,2):1,6,7,8,9         (6,2):1,6,7,8,9         (7,2):1,3,4,5,6,7,8,9   (8,2):1,3,4,5,6,7,8,9   (9,2):1,3,4,5,6,7,8,9
        (1,3):1,3,4,5,6,7,8,9   (2,3):1,3,4,5,6,7,8,9   (3,3):1,3,4,5,6,7,8,9   (4,3):1,2,6,7,8,9       (5,3):1,2,6,7,8,9       (6,3):1,2,6,7,8,9       (7,3):1,2,3,4,5,6,7,8,9 (8,3):1,2,3,4,5,6,7,8,9 (9,3):1,2,3,4,5,6,7,8,9
        (1,4):1,3,4,5,6,7,8,9   (2,4):1,2,3,4,5,6,7,8,9 (3,4):1,2,3,4,5,6,7,8,9 (4,4):1,2,4,5,6,7,8,9   (5,4):1,2,3,5,6,7,8,9   (6,4):1,2,3,4,6,7,8,9   (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):1,3,4,5,6,7,8,9   (2,5):1,2,3,4,5,6,7,8,9 (3,5):1,2,3,4,5,6,7,8,9 (4,5):1,2,4,5,6,7,8,9   (5,5):1,2,3,5,6,7,8,9   (6,5):1,2,3,4,6,7,8,9   (7,5):1,2,3,4,5,6,7,8,9 (8,5):1,2,3,4,5,6,7,8,9 (9,5):1,2,3,4,5,6,7,8,9
        (1,6):1,3,4,5,6,7,8,9   (2,6):1,2,3,4,5,6,7,8,9 (3,6):1,2,3,4,5,6,7,8,9 (4,6):1,2,4,5,6,7,8,9   (5,6):1,2,3,5,6,7,8,9   (6,6):1,2,3,4,6,7,8,9   (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,3,4,5,6,7,8,9   (2,7):1,2,3,4,5,6,7,8,9 (3,7):1,2,3,4,5,6,7,8,9 (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,3,5,6,7,8,9   (6,7):1,2,3,4,6,7,8,9   (7,7):1,2,3,4,5,6,7,8,9 (8,7):1,2,3,4,5,6,7,8,9 (9,7):1,2,3,4,5,6,7,8,9
        (1,8):1,3,4,5,6,7,8,9   (2,8):1,2,3,4,5,6,7,8,9 (3,8):1,2,3,4,5,6,7,8,9 (4,8):1,2,4,5,6,7,8,9   (5,8):1,2,3,5,6,7,8,9   (6,8):1,2,3,4,6,7,8,9   (7,8):1,2,3,4,5,6,7,8,9 (8,8):1,2,3,4,5,6,7,8,9 (9,8):1,2,3,4,5,6,7,8,9
        (1,9):1,3,4,5,6,7,8,9   (2,9):1,2,3,4,5,6,7,8,9 (3,9):1,2,3,4,5,6,7,8,9 (4,9):1,2,4,5,6,7,8,9   (5,9):1,2,3,5,6,7,8,9   (6,9):1,2,3,4,6,7,8,9   (7,9):1,2,3,4,5,6,7,8,9 (8,9):1,2,3,4,5,6,7,8,9 (9,9):1,2,3,4,5,6,7,8,9

           ^ This shows the logic bug.
           (7,1), (8,1) and (9,1) should not contain 2!

        */
        solver.possible.display();
        let pos_index = solver.group.pos(7, 2);
        assert_eq!(
            solver.possible.get(pos_index, 2),
            false,
            "(7,2) contains a 2  [{:?}](not possible)",
            solver.possible.possible[pos_index]
        );

        // solver.possible.display();

        /*
        // Validate the broken possible board.
        for i in 3..9 {
            assert!(!solver.possible.get(solver.possible.pos(1, i), 2), "Possible ({},{}) does not have 2", 1+1, i+1);
            assert!(solver.possible.get(solver.possible.pos(2, i), 2), "Possible ({},{}) does have 2", 2+1, i+1);
            if i > 6 {
                assert!(solver.possible.get(solver.possible.pos(0, i), 2), "Possible ({},{}) does have 2", 0+1, i+1);
            }
        }
        */
        for i in 7..9 {
            assert!(
                !solver.possible.get(solver.possible.pos(i, 2), 2),
                "Possible ({},{}) can not contain 2.",
                2 + 1,
                i + 1
            );
        }

        // Test 2 - pairs

        // let p = solver.possible.get(solver.possible.pos(2,4), 2);

        println!("--- logic_test ---");
    }

    // Running just this test:
    // cargo test -p sudoku logic_finalize_row_test -- --ignored --show-output

    #[ignore = "Finalized not completed (NNY!)"]
    #[test]
    fn finalize_row_test() {
        let mut board = AnyBoard::new(3);
        board.set(3, 2, 3);
        board.set(4, 2, 4);
        board.set(5, 2, 5);
        board.set(6, 1, 3);
        board.set(7, 1, 4);
        board.set(8, 1, 5);
        board.display();
        /*
        ╔═══╦═══╦═══╗
        ║   ║   ║   ║
        ║   ║   ║345║
        ║   ║345║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
        */

        let mut solver = AnySolver::new_from(&board);

        // This shows find_pairs is working...
        solver.possible.display();
        /*
        (1,1):1,2,3,4,5,6,7,8,9 (2,1):1,2,3,4,5,6,7,8,9 (3,1):1,2,3,4,5,6,7,8,9 (4,1):1,2,6,7,8,9       (5,1):1,2,6,7,8,9       (6,1):1,2,6,7,8,9       (7,1):1,2,6,7,8,9       (8,1):1,2,6,7,8,9       (9,1):1,2,6,7,8,9
        (1,2):1,2,6,7,8,9       (2,2):1,2,6,7,8,9       (3,2):1,2,6,7,8,9       (4,2):1,2,6,7,8,9       (5,2):1,2,6,7,8,9       (6,2):1,2,6,7,8,9       (7,2):                  (8,2):                  (9,2):
        (1,3):1,2,6,7,8,9       (2,3):1,2,6,7,8,9       (3,3):1,2,6,7,8,9       (4,3):                  (5,3):                  (6,3):                  (7,3):1,2,6,7,8,9       (8,3):1,2,6,7,8,9       (9,3):1,2,6,7,8,9
         */

        let pairs = solver
            .possible
            .find_pairs(solver.group.group(Groups::Cell, 0));
        solver.possible.display();
        println!("Pairs: {:?}", pairs);
        // let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![0, 1, 2], vec![2, 3, 4])];
        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![3,4,5,6,7,8],vec![0,1,5,6,7,8]),(vec![0, 1, 2], vec![2, 3, 4])];
        assert_eq!(pairs.0, expected);
        /*
        (1,1):3,4,5             (2,1):3,4,5             (3,1):3,4,5             (4,1):1,2,6,7,8,9       (5,1):1,2,6,7,8,9       (6,1):1,2,6,7,8,9       (7,1):1,2,6,7,8,9       (8,1):1,2,6,7,8,9       (9,1):1,2,6,7,8,9
        (1,2):1,2,6,7,8,9       (2,2):1,2,6,7,8,9       (3,2):1,2,6,7,8,9       (4,2):1,2,6,7,8,9       (5,2):1,2,6,7,8,9       (6,2):1,2,6,7,8,9       (7,2):                  (8,2):                  (9,2):
        (1,3):1,2,6,7,8,9       (2,3):1,2,6,7,8,9       (3,3):1,2,6,7,8,9       (4,3):                  (5,3):                  (6,3):                  (7,3):1,2,6,7,8,9       (8,3):1,2,6,7,8,9       (9,3):1,2,6,7,8,9
         */
    }

    // I'm not sure what I was trying to TRY here...

    #[ignore = "Possible doesn't process these correctly. No, not yet!"]
    #[test]
    fn try_this_out() {
        let mut board = AnyBoard::new(3);
        board[(1, 0)] = 2;
        board[(2, 1)] = 3;
        board[(2, 2)] = 5;
        board[(1, 6)] = 3;
        /*
        board.set(1, 0, 2);
        board.set(2, 1, 3);
        board.set(2, 2, 5);
        board.set(1, 6, 3);
        */
        board[(0, 4)] = 4;
        board[(2, 4)] = 2;
        board[(3, 4)] = 3;
        board[(4, 4)] = 5;
        board[(5, 4)] = 7;
        board[(6, 4)] = 8;
        board[(7, 4)] = 9;
        // board.set(0, 4, 4);
        // board.set(2, 4, 2);
        // board.set(3, 4, 3);
        // board.set(4, 4, 5);
        // board.set(5, 4, 7);
        // board.set(6, 4, 8);
        // board.set(7, 4, 9);

        board[(1, 7)] = 5;
        // board.set(1, 7, 5);
        board.display();
        let mut solver = AnySolver::new_from(&board);
        solver.possible.display();

        /*
           ╔═══╦═══╦═══╗
           ║ 2 ║   ║   ║
           ║  3║   ║   ║
           ║  5║   ║   ║
           ╠═══╬═══╬═══╣
           ║P  ║   ║   ║
           ║4 2║357║89 ║
           ║P  ║   ║   ║
           ╠═══╬═══╬═══╣
           ║ 3 ║   ║   ║
           ║ 5 ║   ║   ║
           ║   ║   ║   ║
           ╚═══╩═══╩═══╝
           P = Possible pair (3,5)

        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,4,6,7,8,9       (4,1):1,3,4,5,6,7,8,9   (5,1):1,3,4,5,6,7,8,9   (6,1):1,3,4,5,6,7,8,9   (7,1):1,3,4,5,6,7,8,9   (8,1):1,3,4,5,6,7,8,9   (9,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,4,6,7,8,9       (3,2):                  (4,2):1,2,4,5,6,7,8,9   (5,2):1,2,4,5,6,7,8,9   (6,2):1,2,4,5,6,7,8,9   (7,2):1,2,4,5,6,7,8,9   (8,2):1,2,4,5,6,7,8,9   (9,2):1,2,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,4,6,7,8,9       (3,3):                  (4,3):1,2,3,4,6,7,8,9   (5,3):1,2,3,4,6,7,8,9   (6,3):1,2,3,4,6,7,8,9   (7,3):1,2,3,4,6,7,8,9   (8,3):1,2,3,4,6,7,8,9   (9,3):1,2,3,4,6,7,8,9
        (1,4):1,2,3,5,6,7,8,9   (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9       (4,4):1,2,3,4,5,6,7,8,9 (5,4):1,2,3,4,5,6,7,8,9 (6,4):1,2,3,4,5,6,7,8,9 (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9       (4,5):1,2,3,5,6,7,8,9   (5,5):1,2,3,5,6,7,8,9   (6,5):1,2,3,5,6,7,8,9   (7,5):1,2,3,5,6,7,8,9   (8,5):1,2,3,5,6,7,8,9   (9,5):1,2,3,5,6,7,8,9
        (1,6):1,2,3,5,6,7,8,9   (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9       (4,6):1,2,3,4,5,6,7,8,9 (5,6):1,2,3,4,5,6,7,8,9 (6,6):1,2,3,4,5,6,7,8,9 (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):                  (3,7):1,2,4,6,7,8,9     (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,4,5,6,7,8,9   (6,7):1,2,4,5,6,7,8,9   (7,7):1,2,4,5,6,7,8,9   (8,7):1,2,4,5,6,7,8,9   (9,7):1,2,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):                  (3,8):1,2,4,6,7,8,9     (4,8):1,2,3,4,6,7,8,9   (5,8):1,2,3,4,6,7,8,9   (6,8):1,2,3,4,6,7,8,9   (7,8):1,2,3,4,6,7,8,9   (8,8):1,2,3,4,6,7,8,9   (9,8):1,2,3,4,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,4,6,7,8,9       (3,9):1,2,4,6,7,8,9     (4,9):1,2,3,4,5,6,7,8,9 (5,9):1,2,3,4,5,6,7,8,9 (6,9):1,2,3,4,5,6,7,8,9 (7,9):1,2,3,4,5,6,7,8,9 (8,9):1,2,3,4,5,6,7,8,9 (9,9):1,2,3,4,5,6,7,8,9
        ^ This shows the logic bug.
          (1,4) and (1,6) should only be 3,5.
          (3,7), (3,8), and (3,9) should not contain 2.
        */

        println!("RIGHT HERE:");
        /*
        let g= solver.group.group(Groups::Cell, 3);
        let pairs = solver.possible.find_pairs(g);
        */
        let pairs = solver
            .possible
            .find_pairs(solver.group.group(Groups::Cell, 3));
        println!("Pairs [cell index 3]: {:?}", pairs);
        let g = solver.group.group(Groups::Row, 4); // .clone();
        let pairs = solver.possible.find_pairs(g);
        println!("Pairs [ row index 4]: {:?}", pairs);

        solver.finalize_cell(3);
        solver.possible.display();
    }

    // Running just this test:
    // cargo test -p sudoku logic_test_pairs -- --ignored --show-output

    #[ignore = "Finalized not completed (NNY!)"]
    #[test]
    fn pairs_test() {
        let mut board = AnyBoard::new(3);
        board.set(1, 0, 2);
        board.set(2, 1, 3);
        board.set(2, 2, 5);
        board.set(1, 6, 3);
        board.set(0, 4, 4);
        board.set(1, 7, 5);
        board.display();
        /*
          ╔═══╦═══╦═══╗
          ║ 2 ║   ║   ║
          ║  3║   ║   ║
          ║  5║   ║   ║
          ╠═══╬═══╬═══╣
          ║P  ║   ║   ║
          ║4  ║   ║   ║
          ║P  ║   ║   ║
          ╠═══╬═══╬═══╣
          ║ 3 ║   ║   ║
          ║ 5 ║   ║   ║
          ║   ║   ║   ║
          ╚═══╩═══╩═══╝
          P = Possible pair (3,5)
        */
        let mut solver = AnySolver::new_from(&board);

        solver.possible.display();
        /*
        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,4,6,7,8,9       (4,1):1,3,4,5,6,7,8,9   (5,1):1,3,4,5,6,7,8,9   (6,1):1,3,4,5,6,7,8,9   (7,1):1,3,4,5,6,7,8,9   (8,1):1,3,4,5,6,7,8,9   (9,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,4,6,7,8,9       (3,2):                  (4,2):1,2,4,5,6,7,8,9   (5,2):1,2,4,5,6,7,8,9   (6,2):1,2,4,5,6,7,8,9   (7,2):1,2,4,5,6,7,8,9   (8,2):1,2,4,5,6,7,8,9   (9,2):1,2,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,4,6,7,8,9       (3,3):                  (4,3):1,2,3,4,6,7,8,9   (5,3):1,2,3,4,6,7,8,9   (6,3):1,2,3,4,6,7,8,9   (7,3):1,2,3,4,6,7,8,9   (8,3):1,2,3,4,6,7,8,9   (9,3):1,2,3,4,6,7,8,9
        (1,4):1,2,3,5,6,7,8,9   (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9       (4,4):1,2,3,4,5,6,7,8,9 (5,4):1,2,3,4,5,6,7,8,9 (6,4):1,2,3,4,5,6,7,8,9 (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9       (4,5):1,2,3,5,6,7,8,9   (5,5):1,2,3,5,6,7,8,9   (6,5):1,2,3,5,6,7,8,9   (7,5):1,2,3,5,6,7,8,9   (8,5):1,2,3,5,6,7,8,9   (9,5):1,2,3,5,6,7,8,9
        (1,6):1,2,3,5,6,7,8,9   (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9       (4,6):1,2,3,4,5,6,7,8,9 (5,6):1,2,3,4,5,6,7,8,9 (6,6):1,2,3,4,5,6,7,8,9 (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):                  (3,7):1,2,4,6,7,8,9     (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,4,5,6,7,8,9   (6,7):1,2,4,5,6,7,8,9   (7,7):1,2,4,5,6,7,8,9   (8,7):1,2,4,5,6,7,8,9   (9,7):1,2,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):                  (3,8):1,2,4,6,7,8,9     (4,8):1,2,3,4,6,7,8,9   (5,8):1,2,3,4,6,7,8,9   (6,8):1,2,3,4,6,7,8,9   (7,8):1,2,3,4,6,7,8,9   (8,8):1,2,3,4,6,7,8,9   (9,8):1,2,3,4,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,4,6,7,8,9       (3,9):1,2,4,6,7,8,9     (4,9):1,2,3,4,5,6,7,8,9 (5,9):1,2,3,4,5,6,7,8,9 (6,9):1,2,3,4,5,6,7,8,9 (7,9):1,2,3,4,5,6,7,8,9 (8,9):1,2,3,4,5,6,7,8,9 (9,9):1,2,3,4,5,6,7,8,9
        ^ This shows the logic bug.
          (1,4) and (1,6) should only be 3,5.
          (3,7), (3,8), and (3,9) should not contain 2.
        */

        println!("RIGHT HERE:");
        let g = solver.group.group(Groups::Cell, 3);
        let pairs = solver.possible.find_pairs(g);
        println!("Pairs [cell index 3]: {:?}", pairs);
        // Pairs [cell index 3]: ([([0, 6], [2, 4])], true)

        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![0, 6], vec![2, 4])];
        // Just check expected, not the "found anything to modify" flag.
        assert_eq!(pairs.0, expected);

        // solver.finalize_cell(3);
        solver.possible.display();
        /*
        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,4,6,7,8,9       (4,1):1,3,4,5,6,7,8,9   (5,1):1,3,4,5,6,7,8,9   (6,1):1,3,4,5,6,7,8,9   (7,1):1,3,4,5,6,7,8,9   (8,1):1,3,4,5,6,7,8,9   (9,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,4,6,7,8,9       (3,2):                  (4,2):1,2,4,5,6,7,8,9   (5,2):1,2,4,5,6,7,8,9   (6,2):1,2,4,5,6,7,8,9   (7,2):1,2,4,5,6,7,8,9   (8,2):1,2,4,5,6,7,8,9   (9,2):1,2,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,4,6,7,8,9       (3,3):                  (4,3):1,2,3,4,6,7,8,9   (5,3):1,2,3,4,6,7,8,9   (6,3):1,2,3,4,6,7,8,9   (7,3):1,2,3,4,6,7,8,9   (8,3):1,2,3,4,6,7,8,9   (9,3):1,2,3,4,6,7,8,9
        (1,4):3,5               (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9       (4,4):1,2,3,4,5,6,7,8,9 (5,4):1,2,3,4,5,6,7,8,9 (6,4):1,2,3,4,5,6,7,8,9 (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9       (4,5):1,2,3,5,6,7,8,9   (5,5):1,2,3,5,6,7,8,9   (6,5):1,2,3,5,6,7,8,9   (7,5):1,2,3,5,6,7,8,9   (8,5):1,2,3,5,6,7,8,9   (9,5):1,2,3,5,6,7,8,9
        (1,6):3,5               (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9       (4,6):1,2,3,4,5,6,7,8,9 (5,6):1,2,3,4,5,6,7,8,9 (6,6):1,2,3,4,5,6,7,8,9 (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):                  (3,7):1,2,4,6,7,8,9     (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,4,5,6,7,8,9   (6,7):1,2,4,5,6,7,8,9   (7,7):1,2,4,5,6,7,8,9   (8,7):1,2,4,5,6,7,8,9   (9,7):1,2,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):                  (3,8):1,2,4,6,7,8,9     (4,8):1,2,3,4,6,7,8,9   (5,8):1,2,3,4,6,7,8,9   (6,8):1,2,3,4,6,7,8,9   (7,8):1,2,3,4,6,7,8,9   (8,8):1,2,3,4,6,7,8,9   (9,8):1,2,3,4,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,4,6,7,8,9       (3,9):1,2,4,6,7,8,9     (4,9):1,2,3,4,5,6,7,8,9 (5,9):1,2,3,4,5,6,7,8,9 (6,9):1,2,3,4,5,6,7,8,9 (7,9):1,2,3,4,5,6,7,8,9 (8,9):1,2,3,4,5,6,7,8,9 (9,9):1,2,3,4,5,6,7,8,9

        ^ This shows that find_pairs is working!
        finalize - should see that 2 is only possible in (1,7), (1,8), and (1,9)
        // OK!  find_pairs does work (making the changes) - so I do need to call it, sometime.
        and not in (3,7), (3,8) and (3,9).
         */

        /*
        Pairs: [{1, 4, 8, 6, 7, 0, 5, 2}, {6, 0, 8, 5, 2}, {0, 6}, {}, {0, 6}, {1, 4, 8, 6, 7, 0, 5, 2}, {1, 4, 8, 6, 7, 0, 5, 2}, {1, 4, 8, 6, 7, 0, 5, 2}, {1, 4, 8, 6, 7, 0, 5, 2}]
        3 and 5 are {0,6} !
         */
        // Is validate_board destroying our work?!?
        // YES IT IS!
        // assert!(solver.validate_board());
        // solver.possible.display();
    }

    #[test]
    fn pairs2_test() {
        let mut board = AnyBoard::new(3);
        board.set(1, 0, 2);
        board.set(2, 1, 3);
        board.set(2, 2, 5);
        board.set(1, 6, 3);

        board.set(0, 4, 4);
        board.set(2, 4, 2);
        board.set(3, 4, 3);
        board.set(4, 4, 5);
        board.set(5, 4, 7);
        board.set(6, 4, 8);
        board.set(7, 4, 9);

        board.set(1, 7, 5);
        board.display();
        /*
           ╔═══╦═══╦═══╗
           ║ 2 ║   ║   ║
           ║  3║   ║   ║
           ║  5║   ║   ║
           ╠═══╬═══╬═══╣
           ║P  ║   ║   ║
           ║4 2║357║89 ║
           ║P  ║   ║   ║
           ╠═══╬═══╬═══╣
           ║ 3 ║   ║   ║
           ║ 5 ║   ║   ║
           ║   ║   ║   ║
           ╚═══╩═══╩═══╝
           P = Possible pair (3,5)
        */
        let mut solver = AnySolver::new_from(&board);
        // solver.possible.display();
        /*
        (1,1):1,6,7,8,9         (2,1):                  (3,1):1,4,6,7,8,9       (4,1):1,3,4,5,6,7,8,9   (5,1):1,3,4,5,6,7,8,9   (6,1):1,3,4,5,6,7,8,9   (7,1):1,3,4,5,6,7,8,9   (8,1):1,3,4,5,6,7,8,9   (9,1):1,3,4,5,6,7,8,9
        (1,2):1,6,7,8,9         (2,2):1,4,6,7,8,9       (3,2):                  (4,2):1,2,4,5,6,7,8,9   (5,2):1,2,4,5,6,7,8,9   (6,2):1,2,4,5,6,7,8,9   (7,2):1,2,4,5,6,7,8,9   (8,2):1,2,4,5,6,7,8,9   (9,2):1,2,4,5,6,7,8,9
        (1,3):1,6,7,8,9         (2,3):1,4,6,7,8,9       (3,3):                  (4,3):1,2,3,4,6,7,8,9   (5,3):1,2,3,4,6,7,8,9   (6,3):1,2,3,4,6,7,8,9   (7,3):1,2,3,4,6,7,8,9   (8,3):1,2,3,4,6,7,8,9   (9,3):1,2,3,4,6,7,8,9
        (1,4):1,2,3,5,6,7,8,9   (2,4):1,6,7,8,9         (3,4):1,2,6,7,8,9       (4,4):1,2,3,4,5,6,7,8,9 (5,4):1,2,3,4,5,6,7,8,9 (6,4):1,2,3,4,5,6,7,8,9 (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):                  (2,5):1,6,7,8,9         (3,5):1,2,6,7,8,9       (4,5):1,2,3,5,6,7,8,9   (5,5):1,2,3,5,6,7,8,9   (6,5):1,2,3,5,6,7,8,9   (7,5):1,2,3,5,6,7,8,9   (8,5):1,2,3,5,6,7,8,9   (9,5):1,2,3,5,6,7,8,9
        (1,6):1,2,3,5,6,7,8,9   (2,6):1,6,7,8,9         (3,6):1,2,6,7,8,9       (4,6):1,2,3,4,5,6,7,8,9 (5,6):1,2,3,4,5,6,7,8,9 (6,6):1,2,3,4,5,6,7,8,9 (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):                  (3,7):1,2,4,6,7,8,9     (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,4,5,6,7,8,9   (6,7):1,2,4,5,6,7,8,9   (7,7):1,2,4,5,6,7,8,9   (8,7):1,2,4,5,6,7,8,9   (9,7):1,2,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):                  (3,8):1,2,4,6,7,8,9     (4,8):1,2,3,4,6,7,8,9   (5,8):1,2,3,4,6,7,8,9   (6,8):1,2,3,4,6,7,8,9   (7,8):1,2,3,4,6,7,8,9   (8,8):1,2,3,4,6,7,8,9   (9,8):1,2,3,4,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):1,4,6,7,8,9       (3,9):1,2,4,6,7,8,9     (4,9):1,2,3,4,5,6,7,8,9 (5,9):1,2,3,4,5,6,7,8,9 (6,9):1,2,3,4,5,6,7,8,9 (7,9):1,2,3,4,5,6,7,8,9 (8,9):1,2,3,4,5,6,7,8,9 (9,9):1,2,3,4,5,6,7,8,9
         */

        let pairs = solver
            .possible
            .find_pairs(solver.group.group(Groups::Cell, 3));
        println!("Pairs [cell index 3]: {:?}", pairs);
        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![0, 6], vec![2, 4])];
        // assert_eq!(pairs, (expected, true));
        // Verify we found the pairs -- not that it updated possible or not.
        assert_eq!(pairs.0, expected);

        // This call here to find_pairs, doesn't remove anything.
        // But it does find a pair.

        let g = solver.group.group(Groups::Row, 4);
        let pairs = solver.possible.find_pairs(g);
        println!("Pairs [ row index 4]: {:?}", pairs);
        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![1, 8], vec![0, 5])];
        assert_eq!(pairs, (expected, false));
    }

    // Running just this test:
    // cargo test -p sudoku logic_test_triple -- --ignored --show-output

    #[ignore = "Finalized not completed (NNY!)"]
    #[test]
    fn triple_test() {
        let mut board = AnyBoard::new(3);
        board.set(2, 0, 3);
        board.set(2, 1, 4);
        board.set(2, 2, 5);
        board.set(1, 6, 3);
        board.set(1, 7, 4);
        board.set(1, 8, 5);
        // board.display();
        /*
        ╔═══╦═══╦═══╗
        ║  3║   ║   ║
        ║  4║   ║   ║
        ║  5║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║ 3 ║   ║   ║
        ║ 4 ║   ║   ║
        ║ 5 ║   ║   ║
        ╚═══╩═══╩═══╝
         */

        let mut solver = AnySolver::new_from(&board);
        // solver.possible.display();
        /*
        (1,1):1,2,6,7,8,9       (2,1):1,2,6,7,8,9       (3,1):                  (4,1):1,2,4,5,6,7,8,9   (5,1):1,2,4,5,6,7,8,9   (6,1):1,2,4,5,6,7,8,9   (7,1):1,2,4,5,6,7,8,9   (8,1):1,2,4,5,6,7,8,9   (9,1):1,2,4,5,6,7,8,9
        (1,2):1,2,6,7,8,9       (2,2):1,2,6,7,8,9       (3,2):                  (4,2):1,2,3,5,6,7,8,9   (5,2):1,2,3,5,6,7,8,9   (6,2):1,2,3,5,6,7,8,9   (7,2):1,2,3,5,6,7,8,9   (8,2):1,2,3,5,6,7,8,9   (9,2):1,2,3,5,6,7,8,9
        (1,3):1,2,6,7,8,9       (2,3):1,2,6,7,8,9       (3,3):                  (4,3):1,2,3,4,6,7,8,9   (5,3):1,2,3,4,6,7,8,9   (6,3):1,2,3,4,6,7,8,9   (7,3):1,2,3,4,6,7,8,9   (8,3):1,2,3,4,6,7,8,9   (9,3):1,2,3,4,6,7,8,9
        (1,4):1,2,3,4,5,6,7,8,9 (2,4):1,2,6,7,8,9       (3,4):1,2,6,7,8,9       (4,4):1,2,3,4,5,6,7,8,9 (5,4):1,2,3,4,5,6,7,8,9 (6,4):1,2,3,4,5,6,7,8,9 (7,4):1,2,3,4,5,6,7,8,9 (8,4):1,2,3,4,5,6,7,8,9 (9,4):1,2,3,4,5,6,7,8,9
        (1,5):1,2,3,4,5,6,7,8,9 (2,5):1,2,6,7,8,9       (3,5):1,2,6,7,8,9       (4,5):1,2,3,4,5,6,7,8,9 (5,5):1,2,3,4,5,6,7,8,9 (6,5):1,2,3,4,5,6,7,8,9 (7,5):1,2,3,4,5,6,7,8,9 (8,5):1,2,3,4,5,6,7,8,9 (9,5):1,2,3,4,5,6,7,8,9
        (1,6):1,2,3,4,5,6,7,8,9 (2,6):1,2,6,7,8,9       (3,6):1,2,6,7,8,9       (4,6):1,2,3,4,5,6,7,8,9 (5,6):1,2,3,4,5,6,7,8,9 (6,6):1,2,3,4,5,6,7,8,9 (7,6):1,2,3,4,5,6,7,8,9 (8,6):1,2,3,4,5,6,7,8,9 (9,6):1,2,3,4,5,6,7,8,9
        (1,7):1,2,6,7,8,9       (2,7):                  (3,7):1,2,6,7,8,9       (4,7):1,2,4,5,6,7,8,9   (5,7):1,2,4,5,6,7,8,9   (6,7):1,2,4,5,6,7,8,9   (7,7):1,2,4,5,6,7,8,9   (8,7):1,2,4,5,6,7,8,9   (9,7):1,2,4,5,6,7,8,9
        (1,8):1,2,6,7,8,9       (2,8):                  (3,8):1,2,6,7,8,9       (4,8):1,2,3,5,6,7,8,9   (5,8):1,2,3,5,6,7,8,9   (6,8):1,2,3,5,6,7,8,9   (7,8):1,2,3,5,6,7,8,9   (8,8):1,2,3,5,6,7,8,9   (9,8):1,2,3,5,6,7,8,9
        (1,9):1,2,6,7,8,9       (2,9):                  (3,9):1,2,6,7,8,9       (4,9):1,2,3,4,6,7,8,9   (5,9):1,2,3,4,6,7,8,9   (6,9):1,2,3,4,6,7,8,9   (7,9):1,2,3,4,6,7,8,9   (8,9):1,2,3,4,6,7,8,9   (9,9):1,2,3,4,6,7,8,9
         */

        println!("RIGHT HERE:");
        let g = solver.group.group(Groups::Cell, 3);
        let pairs = solver.possible.find_pairs(g);
        println!("Pairs [cell index 3]: {:?}", pairs);
        // Pairs [cell index 3]: ([([0, 3, 6], [2, 3, 4])], true)
        // [1, 2, 4, 5, 7, 8], [0, 1, 5, 6, 7, 8]
        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![1,2,4,5,7,8],vec![0,1,5,6,7,8]),(vec![0, 3, 6], vec![2, 3, 4])];
        assert_eq!(pairs.0, expected);

        // solver.finalize_cell(3);
        // solver.possible.display();
    }

    /*
    10/13/2024
    This assumes the solver is working correctly.
    I broken something this morning, so this is now hanging.
    No, it just takes a longer time to run (because of debug output,
    and of all of the lengthy calculations I'm doing every time).
     */

    #[test]
    fn display_board() {
        let mut board = AnyBoard::new(3);
        let result = board.load_from_tld(
            'b',
            '_',
            "__c_____e_h__cb___bd___ei_ch_jb__d___________i_eh__b__dg___ij_f_i__jg_____b_____g",
        );
        assert!(result.is_ok());
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "  17 83  ",
                " 73   68 ",
                "2  9 4  1",
                "   1 7   ",
                " 2     9 ",
                " 14   86 ",
                "  83 19  ",
                "         ",
                "4 2   5 6"
            ]
        );

        board.display();
        let mut solver = AnySolver::new_from(&board);
        assert!(solver.validate_board());

        solver.reset_possible();

        if !solver.solve_logic() {
            println!("Failed to solve by logic!");
            solver.board.display();
        }
        solver.board.display();

        /*
        ╔═══╦═══╦═══╗
        ║541║768║329║
        ║973║512║684║
        ║286║934║751║
        ╠═══╬═══╬═══╣
        ║869║157║243║
        ║325║486║197║
        ║714║293║865║
        ╠═══╬═══╬═══╣
        ║658║341║972║
        ║197║625║438║
        ║432║879║516║
        ╚═══╩═══╩═══╝
         */
        assert!(solver.validate_board());
        assert!(solver.board.complete());

        let strings = solver.board.to_strings();
        assert_eq!(
            strings,
            vec![
                "541768329",
                "973512684",
                "286934751",
                "869157243",
                "325486197",
                "714293865",
                "658341972",
                "197625438",
                "432879516"
            ]
        );

        board = AnyBoard::new(4);
        let result = board.load_from_tld('b', '_', "_bo_j_m_f__dp__ge_h_pcfdo___q__n___qio______df___f__l___hpnm_i___obig_p_qhl__k_m_dq_cn______o_g_p_____bi_kc__jn______fo____gi______eb____jd______jk__ml_bn_____i_m_b______oq_nj_d_n__jck_m_fgbq___i_medp___n__b___dg______qjk___j__p___fgohl_d_qo__mq__g_d_p_le_");
        assert!(result.is_ok());
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " D    O    C  IN",
                "A  ENC   IL     ",
                "NG  AP   J MHC  ",
                "  P H   D A  FOL",
                "IOHKFB  A   L  P",
                " BN  M E L ID   ",
                "LE  O AN K BC   ",
                " C    H    JO EF",
                "EN GP    A    F ",
                "   OG J IM L  NC",
                "   MK B C N  PG ",
                "C  L   F  PEMIKO",
                "OPC  N H   F J  ",
                "  EHJ I   MA  CK",
                "     FM   IPA  D",
                "FM  L    H    P "
            ]
        );
        board.display();

        let mut solver = AnySolver::new_from(&board);
        assert!(solver.validate_board());

        if !solver.solve_logic() {
            solver.board.display();
        }
        solver.board.display();

        /*
        solver.board.display();
        ╔════╦════╦════╦════╗
        ║HDLJ║MKOB║FPEC║GAIN║
        ║AKOE║NCFD║GILH║PMBJ║
        ║NGFB║APLI║OJKM║HCDE║
        ║MIPC║HJEG║DBAN║KFOL║
        ╠════╬════╬════╬════╣
        ║IOHK║FBDJ║AECG║LNMP║
        ║JBNF║CMPE║HLOI║DKAG║
        ║LEMD║OGAN║PKFB║CHJI║
        ║GCAP║ILHK║MNDJ║OBEF║
        ╠════╬════╬════╬════╣
        ║ENIG║POCM║KABD║JLFH║
        ║PFKO║GEJA║IMHL║BDNC║
        ║DHJM║KIBL║CFNO║EPGA║
        ║CABL║DHNF║JGPE║MIKO║
        ╠════╬════╬════╬════╣
        ║OPCA║ENKH║BDGF║IJLM║
        ║BLEH║JDIP║NOMA║FGCK║
        ║KJGN║BFMO║LCIP║AEHD║
        ║FMDI║LAGC║EHJK║NOPB║
        ╚════╩════╩════╩════╝
        */
        assert!(solver.validate_board());
        assert!(solver.board.complete());
        let strings: Vec<String> = solver.board.to_strings();
        assert_eq!(
            strings,
            vec![
                "HDLJMKOBFPECGAIN",
                "AKOENCFDGILHPMBJ",
                "NGFBAPLIOJKMHCDE",
                "MIPCHJEGDBANKFOL",
                "IOHKFBDJAECGLNMP",
                "JBNFCMPEHLOIDKAG",
                "LEMDOGANPKFBCHJI",
                "GCAPILHKMNDJOBEF",
                "ENIGPOCMKABDJLFH",
                "PFKOGEJAIMHLBDNC",
                "DHJMKIBLCFNOEPGA",
                "CABLDHNFJGPEMIKO",
                "OPCAENKHBDGFIJLM",
                "BLEHJDIPNOMAFGCK",
                "KJGNBFMOLCIPAEHD",
                "FMDILAGCEHJKNOPB"
            ]
        );

        let mut board: AnyBoard = AnyBoard::new(5);
        let result = board.load_from_tld('b', '_', "_m_kzg____h_s_n____ftd_u_u_dh_____f_b_t_w_____yg_c_rl_o_tdhy__m__uvig_w_sk_eg___p_q_jikouys_r_d___mlq_t_sb_emcwg_dlzyo_kp_i_ng__ir_b_fp_vhz_ce_y_jm__w__m__o_k_xul_qbt_d_s__e____otv_dhegn___mfkpz_blr____s_dv_n_mjx_ckg_w_bo_p___kqyelwjcz_____nxumoisdh_z__fp_vbi_______dkx_eg__r_y_mlwf_u__q_i__o_chdv_j_i_he_r_____________p_zl_k_d_vbjh_y__e_p__s_tguc_q_s__qj_kpn_______ufw_hx__i_hvntirfxw_____lbckympjg___u_kz_m_bfn_yvx_h_ir_o____rgm_otlnx___ipfes_kwc____p__c_v_ugh_krj_m_w__x__x__ci_j_qk_mpo_dr_u_zb__ht_i_qe_wjvcy_bhkzx_ng_u_syv___u_c_hsfrlqo_t_e___pj_cn_h_slzr__j__mqgp_y_vd_m_bs_____t_o_n_h_____ez_f_e_ufd____p_g_z____cqr_x_");
        assert!(result.is_ok());
        board.display();
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " T DPF    Y H R    WSX L ",
                "L QF     J X C G     UB D",
                " CK S LNRP  G  UTQO H MA ",
                "JG   H S XELDUPM F B   RT",
                "Y N RQ UCDOK AISJL HP G E",
                "F  OA N UK VQI HY B DT  C",
                "  S  A C VUE GJQ N I  R  ",
                "  CPD JGMIA   OELSU VBK  ",
                "  G LE D BHT XMW K PI Y  ",
                " EXIBOWFLY     VAMTJUGQS ",
                "G  HV TMI       EWF BR  O",
                " A JFUK W  P D  M GLXE N ",
                "R LN G             O QI F",
                " S TCYP B  H O  X JNAK M ",
                "M  XK ALJ       UHQ GP  Y",
                " VTRYBSEFM     KWOICJNLG ",
                "  U XD J WCN RTA E QY P  ",
                "  HQN COVTJ   EBGDL WSF  ",
                "  F  X Y LWB SVJ R T  O  ",
                "E  CJ R AN GOF XH V MD  B",
                "S V OI ANHDC TGLQJ YF X P",
                "CX   L K RFUYBWO V A   DQ",
                " FR H DQOC  K  INBW T UY ",
                "T JL     G I P F     OC W",
                " B KMV    Q J H    GRI E "
            ]
        );
        let mut solver = AnySolver::new_from(&board);
        assert!(solver.validate_board());

        // This takes awhile to solve...
        if !solver.solve_logic() {
            solver.board.display();
        }
        assert!(solver.validate_board());
        assert!(solver.board.complete());
        solver.board.display();

        /*
        solver.board.display();
        ╔═════╦═════╦═════╦═════╦═════╗
        ║UTODP║FIBGA║YMHQR║NKCEW║SXJLV║
        ║LHQFI║MYTEJ║SXWCN║GVPAR║OUBKD║
        ║XCKES║WLNRP║VFGJB║UTQOD║HYMAI║
        ║JGAVW║HOSKX║ELDUP║MIFYB║QCNRT║
        ║YMNBR║QVUCD║OKTAI║SJLXH║PWGFE║
        ╠═════╬═════╬═════╬═════╬═════╣
        ║FJMOA║RNPUK║LVQIS║HYGBX║DTEWC║
        ║KWSYT║AHCXV║UEBGJ║QONDI║LFRPM║
        ║QRCPD║TJGMI║AYNWO║ELSUF║VBKHX║
        ║VNGUL║EQDSB║HTFXM║WCKRP║IAYOJ║
        ║HEXIB║OWFLY║RDPKC║VAMTJ║UGQSN║
        ╠═════╬═════╬═════╬═════╬═════╣
        ║GQPHV║CTMIS║XJANL║YEWFK║BRDUO║
        ║OABJF║UKVWQ║IPCDY║RMTGL║XESNH║
        ║RYLNU║GXHDE║TWVMK║PBASO║CQIJF║
        ║WSETC║YPRBF║GHUOQ║DXIJN║AKVML║
        ║MIDXK║NALJO║BRSEF║CUHQV║GPWTY║
        ╠═════╬═════╬═════╬═════╬═════╣
        ║AVTRY║BSEFM║PQXHD║KWOIC║JNLGU║
        ║BOUSX║DGJHW║CNLRT║AFEMQ║YVPIK║
        ║IPHQN║KCOVT║JAMYE║BGDLU║WSFXR║
        ║DKFMG║XUYQL║WBISV║JPRNT║EHOCA║
        ║ELWCJ║PRIAN║KGOFU║XHYVS║MDTQB║
        ╠═════╬═════╬═════╬═════╬═════╣
        ║SUVWO║IEANH║DCRTG║LQJKY║FMXBP║
        ║CXIGE║LMKTR║FUYBW║OSVPA║NJHDQ║
        ║PFRAH║JDQOC║MSKVX║INBWE║TLUYG║
        ║TDJLQ║SBXYG║NIEPA║FRUHM║KOCVW║
        ║NBYKM║VFWPU║QOJLH║TDXCG║RIAES║
        ╚═════╩═════╩═════╩═════╩═════╝
         */
        let strings: Vec<String> = solver.board.to_strings();
        assert_eq!(
            strings,
            vec![
                "UTODPFIBGAYMHQRNKCEWSXJLV",
                "LHQFIMYTEJSXWCNGVPAROUBKD",
                "XCKESWLNRPVFGJBUTQODHYMAI",
                "JGAVWHOSKXELDUPMIFYBQCNRT",
                "YMNBRQVUCDOKTAISJLXHPWGFE",
                "FJMOARNPUKLVQISHYGBXDTEWC",
                "KWSYTAHCXVUEBGJQONDILFRPM",
                "QRCPDTJGMIAYNWOELSUFVBKHX",
                "VNGULEQDSBHTFXMWCKRPIAYOJ",
                "HEXIBOWFLYRDPKCVAMTJUGQSN",
                "GQPHVCTMISXJANLYEWFKBRDUO",
                "OABJFUKVWQIPCDYRMTGLXESNH",
                "RYLNUGXHDETWVMKPBASOCQIJF",
                "WSETCYPRBFGHUOQDXIJNAKVML",
                "MIDXKNALJOBRSEFCUHQVGPWTY",
                "AVTRYBSEFMPQXHDKWOICJNLGU",
                "BOUSXDGJHWCNLRTAFEMQYVPIK",
                "IPHQNKCOVTJAMYEBGDLUWSFXR",
                "DKFMGXUYQLWBISVJPRNTEHOCA",
                "ELWCJPRIANKGOFUXHYVSMDTQB",
                "SUVWOIEANHDCRTGLQJKYFMXBP",
                "CXIGELMKTRFUYBWOSVPANJHDQ",
                "PFRAHJDQOCMSKVXINBWETLUYG",
                "TDJLQSBXYGNIEPAFRUHMKOCVW",
                "NBYKMVFWPUQOJLHTDXCGRIAES"
            ]
        );
    }

    #[test]
    fn find_pairs_three_test() {
        let mut board = AnyBoard::new(3);
        board.set(3, 2, 3);
        board.set(4, 2, 4);
        board.set(5, 2, 5);
        board.set(6, 1, 3);
        board.set(7, 1, 4);
        board.set(8, 1, 5);
        board.display();
        /*
        ╔═══╦═══╦═══╗
        ║   ║   ║   ║
        ║   ║   ║345║
        ║   ║345║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
        */

        let mut solver = AnySolver::new_from(&board);

        // This shows find_pairs is working...
        solver.possible.display();
        /*
        (1,1):1,2,3,4,5,6,7,8,9 (2,1):1,2,3,4,5,6,7,8,9 (3,1):1,2,3,4,5,6,7,8,9 (4,1):1,2,6,7,8,9       (5,1):1,2,6,7,8,9       (6,1):1,2,6,7,8,9       (7,1):1,2,6,7,8,9       (8,1):1,2,6,7,8,9       (9,1):1,2,6,7,8,9
        (1,2):1,2,6,7,8,9       (2,2):1,2,6,7,8,9       (3,2):1,2,6,7,8,9       (4,2):1,2,6,7,8,9       (5,2):1,2,6,7,8,9       (6,2):1,2,6,7,8,9       (7,2):                  (8,2):                  (9,2):
        (1,3):1,2,6,7,8,9       (2,3):1,2,6,7,8,9       (3,3):1,2,6,7,8,9       (4,3):                  (5,3):                  (6,3):                  (7,3):1,2,6,7,8,9       (8,3):1,2,6,7,8,9       (9,3):1,2,6,7,8,9
         */

        let pairs = solver
            .possible
            .find_pairs(solver.group.group(Groups::Cell, 0));
        // solver.possible.display();
        /*
        (1,1):3,4,5             (2,1):3,4,5             (3,1):3,4,5             (4,1):1,2,6,7,8,9       (5,1):1,2,6,7,8,9       (6,1):1,2,6,7,8,9       (7,1):1,2,6,7,8,9       (8,1):1,2,6,7,8,9       (9,1):1,2,6,7,8,9
        (1,2):1,2,6,7,8,9       (2,2):1,2,6,7,8,9       (3,2):1,2,6,7,8,9       (4,2):1,2,6,7,8,9       (5,2):1,2,6,7,8,9       (6,2):1,2,6,7,8,9       (7,2):                  (8,2):                  (9,2):
        (1,3):1,2,6,7,8,9       (2,3):1,2,6,7,8,9       (3,3):1,2,6,7,8,9       (4,3):                  (5,3):                  (6,3):                  (7,3):1,2,6,7,8,9       (8,3):1,2,6,7,8,9       (9,3):1,2,6,7,8,9
         */

        // println!("Pairs: {:?}", pairs);
        // Actually, this is wrong:
        // let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![(vec![0, 1, 2], vec![2, 3, 4])];
        // assert_eq!(pairs, (expected, true));
        // There are pairs (of 6 values) in (4,1)-(9,1)!
        let expected: Vec<(Vec<usize>, Vec<u8>)> = vec![
            (vec![3, 4, 5, 6, 7, 8], vec![0, 1, 5, 6, 7, 8]),
            (vec![0, 1, 2], vec![2, 3, 4]),
        ];
        // don't verify that find_pairs updated anything...
        assert_eq!(pairs.0, expected);
    }

    #[test]
    fn solve_board() {
        let mut board = AnyBoard::new(3);
        let result = board.load_from_tld(
            'b',
            '_',
            "__c_____e_h__cb___bd___ei_ch_jb__d___________i_eh__b__dg___ij_f_i__jg_____b_____g",
        );
        assert!(result.is_ok());
        assert_eq!(3, board.size);
        assert_eq!(9, board.width);
        assert_eq!(81, board.max_index);

        // board.display();
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "  17 83  ",
                " 73   68 ",
                "2  9 4  1",
                "   1 7   ",
                " 2     9 ",
                " 14   86 ",
                "  83 19  ",
                "         ",
                "4 2   5 6"
            ]
        );

        let mut solver = AnySolver::new_from(&board);
        assert!(solver.validate_board());
        let solutions = solver.board.brute_force_solver(2);
        assert!(solutions.0 == 1, "Expected 1 solution, got {}", solutions.0);
        let strings: Vec<String> = solutions.1[0].to_strings();
        assert_eq!(
            strings,
            vec![
                "541768329",
                "973512684",
                "286934751",
                "869157243",
                "325486197",
                "714293865",
                "658341972",
                "197625438",
                "432879516",
            ]
        );

        // 4x4 board takes 40 minutes

        if false {
            board = AnyBoard::new(4);
            let result = board.load_from_tld('b', '_', "_bo_j_m_f__dp__ge_h_pcfdo___q__n___qio______df___f__l___hpnm_i___obig_p_qhl__k_m_dq_cn______o_g_p_____bi_kc__jn______fo____gi______eb____jd______jk__ml_bn_____i_m_b______oq_nj_d_n__jck_m_fgbq___i_medp___n__b___dg______qjk___j__p___fgohl_d_qo__mq__g_d_p_le_");
            assert!(result.is_ok());
            board.display();

            let mut solver = AnySolver::new_from(&board);
            assert!(solver.validate_board());
            let solutions = solver.board.brute_force_solver(2);
            assert!(solutions.0 == 1);
        }
    }

    #[test]
    fn make_board() {
        // Making a 4x4 board varies (0.3-60 seconds).
        // Maybe we don't want to do this in a test, because of the randomness involved.

        let mut solver = AnySolver::new(3);
        solver.make();
        solver.board.display();
        assert!(solver.validate_board());
    }

    #[test]
    fn validated_board() {
        // Create an invalid board. Position (5,9) can't possibly hold any value.

        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(4, 1, 2);
        board.set(4, 2, 3);
        board.set(4, 3, 4);
        board.set(4, 4, 5);
        board.set(4, 5, 6);
        board.set(0, 8, 7);
        board.set(1, 8, 8);
        board.set(2, 8, 9);
        /*
        // board.display();
        ╔═══╦═══╦═══╗
        ║   ║ 1 ║   ║
        ║   ║ 2 ║   ║
        ║   ║ 3 ║   ║
        ╠═══╬═══╬═══╣
        ║   ║ 4 ║   ║
        ║   ║ 5 ║   ║
        ║   ║ 6 ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║789║ ! ║   ║
        ╚═══╩═══╩═══╝
        The ! can't possibly be anything.
         */

        let mut solver = AnySolver::new_from(&board);
        assert!(!solver.validate_board());

        // Invalid board: Has two 1's in same column & cell.
        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(4, 1, 1);
        /*
        // board.display();
        ╔═══╦═══╦═══╗
        ║   ║ 1 ║   ║
        ║   ║ 1 ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╠═══╬═══╬═══╣
        ║   ║   ║   ║
        ║   ║   ║   ║
        ║   ║   ║   ║
        ╚═══╩═══╩═══╝
        */

        // Because of the new AnySolver.process_move code --
        // This panics (as would validate_board!)

        let mut solver = AnySolver::new_from(&board);

        assert!(!solver.validate_board());

        // Invalid board: Has two 1's in same row & cell.
        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(5, 0, 1);
        // board.display();

        // Invalid board: Has two 1's in same column.
        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(4, 4, 1);
        // board.display();

        let mut solver = AnySolver::new_from(&board);
        assert!(!solver.validate_board());

        // Invalid board: Has two 1's in same row.
        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(7, 0, 1);
        // board.display();
        let mut solver = AnySolver::new_from(&board);
        assert!(!solver.validate_board());

        // Invalid board: Has two 1's in same cell.
        let mut board = AnyBoard::new(3);
        board.set(4, 0, 1);
        board.set(5, 1, 1);
        // board.display();

        let mut solver = AnySolver::new_from(&board);
        assert!(!solver.validate_board());
    }
}
